"""This is an example of how to use TEEHR in NGEN."""
from teehr import Evaluation
from pathlib import Path
import shutil
from teehr import Metrics as metrics


# Set a path to the directory where the evaluation will be created
TEST_STUDY_DIR = Path(Path().home(), "temp", "ngen_example")
shutil.rmtree(TEST_STUDY_DIR, ignore_errors=True)
TEST_STUDY_DIR.mkdir(parents=True, exist_ok=True)

# Set paths to the locations and crosswalks
# In NGEN this will be provided by NGEN.
TEST_DATA = "/home/matt/repos/teehr/tests/data/two_locations/"
LOCATIONS = Path(TEST_DATA, "two_locations.parquet")
LOCATION_XWALKS = Path(TEST_DATA, "two_crosswalks.parquet")
STARTDATE = "2010-10-01"
ENDDATE = "2020-09-30"

# Create an Evaluation object
ev = Evaluation(dir_path=TEST_STUDY_DIR)

# Enable logging
ev.enable_logging()

# Clone the template
ev.clone_template()

# Load the location data (observations)
# This is a geoparquet file, but any format
# GeoPandas supports is supported.
ev.locations.load_spatial(in_path=LOCATIONS)

# Load the crosswalk data for USGS to NWM3.0
# Crosswalks are needed for each "simulation source"
# i.e. USGS -> NWM3.0, USGS -> NGEN, etc.
ev.location_crosswalks.load_parquet(
    in_path=LOCATION_XWALKS
)

# =====>>> This is where USGS -> NGEN crosswalk will be loaded. <<<=====
# ev.load.import_location_crosswalks()

# Load the USGS observations
# Note, dates are hard coded and will be provided by NGEN
ev.fetch.usgs_streamflow(
    start_date=STARTDATE,
    end_date=ENDDATE
)

# Load the NWM retrospective data
ev.fetch.nwm_retrospective_points(
    nwm_version="nwm30",
    variable_name="streamflow",
    start_date=STARTDATE,
    end_date=ENDDATE
)

# =====>>> This is where NGEN simulation output will be loaded. <<<=====
# ev.load.import_secondary_timeseries()

# Create the joined timeseries
ev.joined_timeseries.create(execute_scripts=False)

df = ev.metrics.query(
    order_by=["primary_location_id", "configuration_name"],
    group_by=["primary_location_id", "configuration_name"],
    include_metrics=[
        metrics.KlingGuptaEfficiency(),
        metrics.NashSutcliffeEfficiency(),
        metrics.RelativeBias()
    ]
).to_pandas()

df.to_csv(Path(TEST_STUDY_DIR, "metrics.csv"), index=False)



"""Classes for bootstrapping sampling methods."""
from typing import Callable, List, Union
from pathlib import Path

from numpy.random import RandomState
from pydantic import Field

import teehr.metrics.bootstrap_funcs as bootstrap_funcs
from teehr.models.metrics.basemodels import BootstrapBasemodel


class Gumboot(BootstrapBasemodel):
    """Gumboot bootstrapping.

    This is a partial implementation of the Gumboot R package, a
    non-overlapping bootstrap method where blocks are defined by water
    years. Synthetic timeseries are constructed by randomly resampling water
    years from the input timeseries with replacement. The specified performance
    metric is calculated for each synthetic timeseries for a number of
    bootstrap replications (reps). The quantiles of the bootstrap metric
    results are calculated and returned.

    If the quantile values are not specified or are set to None, the array
    of metric values is returned (dimensions: [reps, 1]). Otherwise the
    specified quantiles of the metric values are returned as a dictionary.

    See Also:  Clark et al. (2021), "The abuse of popular performance metrics
      in hydrologic modeling", Water Resources Research,
      <doi:10.1029/2020WR029001>

      https://cran.r-project.org/web/packages/gumboot/gumboot.pdf

    Parameters
    ----------
    reps : int
        The number of bootstrap replications. Default value is 1000.
    seed : Union[int, None]
        The seed for the random number generator. Setting a seed value can be
        used to provide reproducible results. Default value is None.
    quantiles : Union[List[float], None]
        The quantiles to calculate from the bootstrap metric results. The
        default value is None.
    boot_year_file : Union[str, Path, None]
        The file path to the boot year csv file. The default value is None.
    water_year_month : int
        The month specifying the start of the water year. Default value is 10.
    name : str
        The name of the bootstrap method. Currently only used in
        logging. Default value is "Gumboot".
    include_value_time : bool, fixed
        Whether to include the value_time series in the bootstrapping
        function. Default value is True.
    func : Callable, fixed
        The wrapper to generate the bootstrapping function. Default value is
        bootstrap_funcs.create_gumboot_func.

    """

    reps: int = 1000
    seed: Union[int, None] = None
    quantiles: Union[List[float], None] = None
    boot_year_file: Union[str, Path, None] = None
    water_year_month: int = 10
    name: str = Field(default="Gumboot")
    include_value_time: bool = Field(True, frozen=True)
    func: Callable = Field(bootstrap_funcs.create_gumboot_func, frozen=True)


class CircularBlock(BootstrapBasemodel):
    """CircularBlock bootstrapping from the arch python package.

    Parameters
    ----------
    seed : Union[int, None]
        The seed for the random number generator. Default value is None.
    random_state : RandomState, optional
        The random state for the random number generator.
    reps : int
        The number of bootstrap replications.
    block_size : int
        The block size for the CircularBlockBootstrap.
    quantiles : List[float]
        The quantiles to calculate from the bootstrap results. Default
        value is None.
    name : str
        The name of the bootstrap method. Currently only used in
        logging. Default value is "CircularBlock".
    include_value_time : bool, fixed
        Whether to include the value_time series in the bootstrapping
        function. Default value is True.
    func : Callable, fixed
        The wrapper to generate the bootstrapping function.
    """

    seed: Union[int, None] = None
    random_state: Union[RandomState, None] = None
    reps: int = 1000
    block_size: int = 365
    quantiles: Union[List[float], None] = None
    name: str = Field(default="CircularBlock")
    include_value_time: bool = Field(False, frozen=True)
    func: Callable = Field(
        bootstrap_funcs.create_circularblock_func,
        frozen=True
    )


class Stationary(BootstrapBasemodel):
    """Stationary bootstrapping from the arch python package.

    Parameters
    ----------
    seed : Union[int, None]
        The seed for the random number generator. Default value is 42.
    random_state : RandomState, optional
        The random state for the random number generator.
    reps : int
        The number of bootstrap replications.
    block_size : int
        The block size for the StationaryBootstrap.
    quantiles : List[float]
        The quantiles to calculate from the bootstrap results. Default
        value is None.
    name : str
        The name of the bootstrap method. Currently only used in
        logging. Default value is "Stationary".
    include_value_time : bool, fixed
        Whether to include the value_time series in the bootstrapping
        function. Default value is True.
    func : Callable, fixed
        The wrapper to generate the bootstrapping function.
    """

    seed: Union[int, None] = None
    random_state: Union[RandomState, None] = None
    reps: int = 1000
    block_size: int = 365
    quantiles: Union[List[float], None] = None
    name: str = Field(default="Stationary")
    include_value_time: bool = Field(False, frozen=True)
    func: Callable = Field(
        bootstrap_funcs.create_stationary_func,
        frozen=True
    )


class Bootstrappers:
    """Container class for bootstrap sampling classes.

    Notes
    -----
    Bootstrapping is a resampling method used to estimate uncertainty
    in metric results. The bootstrapping methods available in TEEHR
    include:

    - Gumboot
    - CircularBlock
    - Stationary
    """

    Gumboot = Gumboot
    CircularBlock = CircularBlock
    Stationary = Stationary



"""Module defining shared functions for processing NWM grid data."""
from pathlib import Path
from typing import Dict, List, Tuple, Union

import dask
import numpy as np
import pandas as pd
import xarray as xr

import teehr.models.pandera_dataframe_schemas as schemas
from teehr.fetching.utils import (
    get_dataset,
    write_timeseries_parquet_file,
    parse_nwm_json_paths,
    format_nwm_configuration_name
)
from teehr.models.fetching.utils import TimeseriesTypeEnum
from teehr.fetching.const import (
    VALUE,
    VALUE_TIME,
    REFERENCE_TIME,
    LOCATION_ID,
    UNIT_NAME,
    VARIABLE_NAME,
    CONFIGURATION_NAME
)


def get_weights_row_col_stats(weights_df: pd.DataFrame) -> Dict:
    """Get row and column statistics for weights dataframe."""
    row_min = weights_df.row.values.min()
    col_min = weights_df.col.values.min()
    row_max = weights_df.row.values.max()
    col_max = weights_df.col.values.max()

    rows_norm = weights_df.row.values - row_min
    cols_norm = weights_df.col.values - col_min
    return {
        "row_min": row_min,
        "row_max": row_max,
        "col_min": col_min,
        "col_max": col_max,
        "rows_norm": rows_norm,
        "cols_norm": cols_norm
    }


def get_nwm_grid_data(
    var_da: xr.DataArray,
    row_min: int,
    col_min: int,
    row_max: int,
    col_max: int
):
    """Read a subset nwm grid data into memory using row/col bounds."""
    grid_values = var_da.isel(
        x=slice(col_min, col_max+1), y=slice(row_min, row_max+1)
    ).values
    return grid_values


def update_location_id_prefix(
    df: pd.DataFrame,
    new_prefix: str
) -> pd.DataFrame:
    """Replace or add the location_id prefix in a dataframe."""
    df = df.copy()
    tmp_df = df.location_id.str.split("-", expand=True)

    df[LOCATION_ID] = df[LOCATION_ID].astype(str)

    if tmp_df.columns.size == 1:
        df.loc[:, 'location_id'] = new_prefix + "-" + df['location_id']
    elif tmp_df.columns.size == 2:
        df.loc[:, 'location_id'] = new_prefix + "-" + tmp_df[1]
    else:
        raise ValueError("Location ID has more than two parts!")

    return df


def compute_weighted_average(
    grid_values: np.ndarray,
    weights_df: pd.DataFrame
) -> pd.DataFrame:
    """Compute weighted average of pixels for given zones and weights."""
    weights_df.loc[:, "weighted_value"] = grid_values * \
        weights_df.weight.values

    # Compute weighted average
    df = weights_df.groupby(
        by=LOCATION_ID, as_index=False)[["weighted_value", "weight"]].sum()
    df.loc[:, VALUE] = df.weighted_value/df.weight

    return df[[LOCATION_ID, VALUE]].copy()


def read_and_validate_weights_file(
    weights_filepath: str
) -> pd.DataFrame:
    """Read weights file from parquet, validating data types."""
    schema = schemas.weights_file_schema()
    weights_df = pd.read_parquet(
        weights_filepath, columns=list(schema.columns.keys())
    )
    return schema.validate(weights_df)


@dask.delayed
def process_single_nwm_grid_file(
    row: Tuple,
    configuration_name: str,
    variable_name: str,
    weights_filepath: str,
    ignore_missing_file: bool,
    location_id_prefix: Union[str, None],
    variable_mapper: Dict[str, Dict[str, str]]
) -> pd.DataFrame:
    """Fetch data for a single reference file and compute weighted average."""
    ds = get_dataset(
        row.filepath,
        ignore_missing_file,
        target_options={'anon': True}
    )
    if not ds:
        return None
    yrmoday = row.day
    z_hour = row.z_hour[1:3]
    ref_time = pd.to_datetime(yrmoday) \
        + pd.to_timedelta(int(z_hour), unit="h")

    nwm_units = ds[variable_name].attrs["units"]
    value_time = ds.time.values[0]
    da = ds[variable_name][0]

    weights_df = read_and_validate_weights_file(weights_filepath)

    weights_bounds = get_weights_row_col_stats(weights_df)

    grid_arr = get_nwm_grid_data(
        da,
        weights_bounds["row_min"],
        weights_bounds["col_min"],
        weights_bounds["row_max"],
        weights_bounds["col_max"]
    )

    grid_values = grid_arr[
        weights_bounds["rows_norm"],
        weights_bounds["cols_norm"]
    ]

    # Calculate mean areal value of selected variable
    df = compute_weighted_average(grid_values, weights_df)

    if not variable_mapper:
        df.loc[:, UNIT_NAME] = nwm_units
        df.loc[:, VARIABLE_NAME] = variable_name
    else:
        df.loc[:, UNIT_NAME] = variable_mapper[UNIT_NAME].\
            get(nwm_units, nwm_units)
        df.loc[:, VARIABLE_NAME] = variable_mapper[VARIABLE_NAME].\
            get(variable_name, variable_name)

    df.loc[:, VALUE_TIME] = value_time
    df.loc[:, REFERENCE_TIME] = ref_time
    df.loc[:, CONFIGURATION_NAME] = configuration_name

    if location_id_prefix:
        df = update_location_id_prefix(df, location_id_prefix)

    return df


def fetch_and_format_nwm_grids(
    json_paths: List[str],
    nwm_configuration_name: str,
    nwm_version: str,
    variable_name: str,
    output_parquet_dir: str,
    zonal_weights_filepath: str,
    ignore_missing_file: bool,
    overwrite_output: bool,
    location_id_prefix: Union[str, None],
    variable_mapper: Dict[str, Dict[str, str]],
    timeseries_type: TimeseriesTypeEnum
):
    """Compute weighted average, grouping by reference time.

    Group a list of json files by reference time and compute the weighted
    average of the variable values for each zone. The results are saved to
    parquet files using TEEHR data model.
    """
    output_parquet_dir = Path(output_parquet_dir)
    if not output_parquet_dir.exists():
        output_parquet_dir.mkdir(parents=True)

    # Format file list into a dataframe and group by reference time
    df_refs = parse_nwm_json_paths(
        json_paths=json_paths
    )

    gps = df_refs.groupby(["day", "z_hour"])

    teehr_config = format_nwm_configuration_name(
        nwm_configuration_name=nwm_configuration_name,
        nwm_version=nwm_version
    )

    for gp in gps:
        _, df = gp

        results = []
        for row in df.itertuples():
            results.append(
                process_single_nwm_grid_file(
                    row,
                    teehr_config["configuration_name"],
                    variable_name,
                    zonal_weights_filepath,
                    ignore_missing_file,
                    location_id_prefix,
                    variable_mapper
                )
            )

        output = dask.compute(*results)

        output = [df for df in output if df is not None]
        if len(output) == 0:
            raise FileNotFoundError("No NWM files for specified input"
                                    "configuration were found in GCS!")
        z_hour_df = pd.concat(output)

        if timeseries_type == TimeseriesTypeEnum.secondary:
            z_hour_df.loc[:, "member"] = teehr_config["member"]

        # Save to parquet.
        yrmoday = df.day.iloc[0]
        z_hour = df.z_hour.iloc[0][1:3]
        ref_time_str = f"{yrmoday}T{z_hour}"
        parquet_filepath = Path(
            Path(output_parquet_dir), f"{ref_time_str}.parquet"
        )
        z_hour_df.sort_values([LOCATION_ID, VALUE_TIME], inplace=True)
        write_timeseries_parquet_file(
            filepath=parquet_filepath,
            overwrite_output=overwrite_output,
            data=z_hour_df,
            timeseries_type=timeseries_type
        )



"""A module for cloning and optionally subsetting evaluations from s3."""
from pathlib import Path
import yaml
import fsspec
import pandas as pd
import teehr.const as const
from datetime import datetime
from typing import Union, Literal, List
import logging
import pyspark

logger = logging.getLogger(__name__)


def list_s3_evaluations(
        format: Literal["pandas", "list"] = "pandas"
) -> Union[list, pd.DataFrame]:
    """List the evaluations in s3.

    Parameters
    ----------
    format : str
        The format of the output. Either "pandas" or "list".
        Default is "pandas".

    """
    logger.info(f"Getting evaluations from s3: {const.S3_EVALUATIONS_PATH}")
    # Read the content of the file using fsspec
    with fsspec.open(const.S3_EVALUATIONS_PATH, 'r', anon=True) as file:
        yaml_content = file.read()

    # Load the YAML content into a dictionary
    yaml_dict = yaml.safe_load(yaml_content)

    if format == "pandas":
        return pd.DataFrame(yaml_dict["evaluations"])
    return yaml_dict["evaluations"]


def subset_the_table(
    ev,
    table,
    sdf_in: pyspark.sql.DataFrame,
    primary_location_ids: Union[None, List[str]],
    start_date: Union[str, datetime, None],
    end_date: Union[str, datetime, None],
) -> pyspark.sql.DataFrame:
    """Subset the dataset based on location and start/end time."""
    if table.name == "locations" and primary_location_ids is not None:
        sdf_in = sdf_in.filter(sdf_in.id.isin(primary_location_ids))
    elif table.name == "location_attributes" and primary_location_ids is not None:
        sdf_in = sdf_in.filter(sdf_in.location_id.isin(primary_location_ids))
    elif table.name == "location_crosswalks" and primary_location_ids is not None:
        sdf_in = sdf_in.filter(
            sdf_in.primary_location_id.isin(primary_location_ids)
        )
    elif table.name == "primary_timeseries":
        if primary_location_ids is not None:
            sdf_in = sdf_in.filter(
                sdf_in.location_id.isin(primary_location_ids)
            )
    elif table.name == "secondary_timeseries":
        if primary_location_ids is not None:
            secondary_ids = (
                ev.location_crosswalks.to_sdf()
                .select("secondary_location_id").rdd.flatMap(lambda x: x).collect()
            )
            sdf_in = sdf_in.filter(sdf_in.location_id.isin(secondary_ids))
    elif table.name == "joined_timeseries":
        if primary_location_ids is not None:
            sdf_in = sdf_in.filter(
                sdf_in.primary_location_id.isin(primary_location_ids)
            )
    if (
        table.name == "primary_timeseries"
        or table.name == "secondary_timeseries"
        or table.name == "joined_timeseries"
    ):
        if start_date is not None:
            sdf_in = sdf_in.filter(sdf_in.value_time >= start_date)
        if end_date is not None:
            sdf_in = sdf_in.filter(sdf_in.value_time <= end_date)

    return sdf_in


def clone_from_s3(
    ev,
    evaluation_name: str,
    primary_location_ids: Union[None, List[str]],
    start_date: Union[str, datetime, None],
    end_date: Union[str, datetime, None],
):
    """Clone an evaluation from s3.

    Copies the evaluation from s3 to the local directory.
    Includes the following tables:
        - units
        - variables
        - attributes
        - configurations
        - locations
        - location_attributes
        - location_crosswalks
        - primary_timeseries
        - secondary_timeseries
        - joined_timeseries

    Also includes the user_defined_fields.py script.

    Parameters
    ----------
    ev : Evaluation
        The Evaluation object.
    evaluation_name : str
        The name of the evaluation to clone.

    Note: future version will allow subsetting the tables to clone.
    """

    # Make the Evaluation directories
    logger.info(f"Creating directories for evaluation: {evaluation_name}")
    Path(ev.cache_dir).mkdir()
    Path(ev.dataset_dir).mkdir()
    Path(ev.scripts_dir).mkdir()

    logger.info(f"Cloning evaluation from s3: {evaluation_name}")
    tables = [
        {
            "table": ev.units
        },
        {
            "table": ev.variables
        },
        {
            "table": ev.attributes
        },
        {
            "table": ev.configurations
        },
        {
            "table": ev.locations
        },
        {
            "table": ev.location_attributes
        },
        {
            "table": ev.location_crosswalks
        },
        {
            "table": ev.primary_timeseries
        },
        {
            "table": ev.secondary_timeseries
        },
        {
            "table": ev.joined_timeseries
        },
    ]

    protocol_list = list_s3_evaluations(format="list")
    protocol = [p for p in protocol_list if p["name"] == evaluation_name][0]
    url = protocol["url"]
    url = url.replace("s3://", "s3a://")
    s3_dataset_path = f"{url}/dataset"

    logger.info(f"Cloning evaluation from s3: {s3_dataset_path}")

    for table in tables:
        table = table["table"]

        logger.debug(f"Making directory {table.dir}")
        Path(table.dir).mkdir()

        logger.debug(f"Cloning {table.name} from {s3_dataset_path}/{table.name}/ to {table.dir}")

        use_table_schema = True
        if table.name == "joined_timeseries":
            use_table_schema = False

        sdf_in = table._read_files(
            path=f"{s3_dataset_path}/{table.name}/",
            use_table_schema=use_table_schema
        )

        sdf_in = subset_the_table(
            ev=ev,
            sdf_in=sdf_in,
            table=table,
            primary_location_ids=primary_location_ids,
            start_date=start_date,
            end_date=end_date
        )

        if table.name == "joined_timeseries":
            table._write_spark_df(sdf_in)
        else:
            table._write_spark_df(sdf_in, write_mode="overwrite")

    # copy scripts path to ev.scripts_dir
    source = f"{url}/scripts/user_defined_fields.py"
    dest = f"{ev.scripts_dir}/user_defined_fields.py"
    logger.debug(f"Copying from {source}/ to {dest}")

    # ToDo: there is a permission issue that prevents copying the entire directory.
    # This works for now.
    with fsspec.open(source, 'r', anon=True) as file:
        with open(dest, 'w') as f:
            f.write(file.read())



"""Timeseries table base class."""
from teehr.evaluation.tables.base_table import BaseTable
from teehr.loading.utils import (
    validate_input_is_xml,
    validate_input_is_csv,
    validate_input_is_netcdf,
    validate_input_is_parquet
)
from teehr.models.filters import TimeseriesFilter
from teehr.querying.utils import join_geometry
from teehr.models.table_enums import TableWriteEnum

from pathlib import Path
from typing import Union

import logging

logger = logging.getLogger(__name__)


class TimeseriesTable(BaseTable):
    """Access methods to timeseries table."""

    def __init__(self, ev):
        """Initialize class."""
        super().__init__(ev)
        self.format = "parquet"
        self.partition_by = [
            "configuration_name",
            "variable_name",
            "reference_time"
        ]
        self.filter_model = TimeseriesFilter
        self.unique_column_set = [
            "location_id",
            "value_time",
            "reference_time",
            "variable_name",
            "unit_name",
        ]

    def to_pandas(self):
        """Return Pandas DataFrame for Primary Timeseries."""
        self._check_load_table()
        df = self.df.toPandas()
        df.attrs['table_type'] = self.name
        df.attrs['fields'] = self.fields()
        return df

    def to_geopandas(self):
        """Return GeoPandas DataFrame."""
        self._check_load_table()
        return join_geometry(self.df, self.ev.locations.to_sdf())

    def load_parquet(
        self,
        in_path: Union[Path, str],
        pattern: str = "**/*.parquet",
        field_mapping: dict = None,
        constant_field_values: dict = None,
        location_id_prefix: str = None,
        write_mode: TableWriteEnum = "append",
        **kwargs
    ):
        """Import primary timeseries parquet data.

        Parameters
        ----------
        in_path : Union[Path, str]
            Path to the timeseries data (file or directory) in
            parquet file format.
        field_mapping : dict, optional
            A dictionary mapping input fields to output fields.
            Format: {input_field: output_field}
        constant_field_values : dict, optional
            A dictionary mapping field names to constant values.
            Format: {field_name: value}
        location_id_prefix : str, optional
            The prefix to add to location IDs.
            Used to ensure unique location IDs across configurations.
            Note, the methods for fetching USGS and NWM data automatically
            prefix location IDs with "usgs" or the nwm version
            ("nwm12, "nwm21", "nwm22", or "nwm30"), respectively.
        write_mode : TableWriteEnum, optional (default: "append")
            The write mode for the table.
            Options are "append", "upsert", and "overwrite".
            If "append", the table will be appended with new data that does
            already exist.
            If "upsert", existing data will be replaced and new data that
            does not exist will be appended.
            If "overwrite", existing partitions receiving new data are
            overwritten.
        **kwargs
            Additional keyword arguments are passed to pd.read_parquet().

        Includes validation and importing data to database.

        Notes
        -----
        The TEEHR Timeseries table schema includes fields:

        - reference_time
        - value_time
        - configuration_name
        - unit_name
        - variable_name
        - value
        - location_id
        """
        logger.info(f"Loading primary timeseries parquet data: {in_path}")

        validate_input_is_parquet(in_path)
        self._load(
            in_path=in_path,
            pattern=pattern,
            field_mapping=field_mapping,
            constant_field_values=constant_field_values,
            location_id_prefix=location_id_prefix,
            write_mode=write_mode,
            **kwargs
        )
        self._load_table()

    def load_csv(
        self,
        in_path: Union[Path, str],
        pattern: str = "**/*.csv",
        field_mapping: dict = None,
        constant_field_values: dict = None,
        location_id_prefix: str = None,
        write_mode: TableWriteEnum = "append",
        **kwargs
    ):
        """Import primary timeseries csv data.

        Parameters
        ----------
        in_path : Union[Path, str]
            Path to the timeseries data (file or directory) in
            csv file format.
        field_mapping : dict, optional
            A dictionary mapping input fields to output fields.
            Format: {input_field: output_field}
        constant_field_values : dict, optional
            A dictionary mapping field names to constant values.
            Format: {field_name: value}
        location_id_prefix : str, optional
            The prefix to add to location IDs.
            Used to ensure unique location IDs across configurations.
            Note, the methods for fetching USGS and NWM data automatically
            prefix location IDs with "usgs" or the nwm version
            ("nwm12, "nwm21", "nwm22", or "nwm30"), respectively.
        write_mode : TableWriteEnum, optional (default: "append")
            The write mode for the table.
            Options are "append", "upsert", and "overwrite".
            If "append", the table will be appended with new data that does
            already exist.
            If "upsert", existing data will be replaced and new data that
            does not exist will be appended.
            If "overwrite", existing partitions receiving new data are overwritten
        **kwargs
            Additional keyword arguments are passed to pd.read_csv().

        Includes validation and importing data to database.

        Notes
        -----
        The TEEHR Timeseries table schema includes fields:

        - reference_time
        - value_time
        - configuration_name
        - unit_name
        - variable_name
        - value
        - location_id
        """
        logger.info(f"Loading primary timeseries csv data: {in_path}")

        validate_input_is_csv(in_path)
        self._load(
            in_path=in_path,
            pattern=pattern,
            field_mapping=field_mapping,
            constant_field_values=constant_field_values,
            location_id_prefix=location_id_prefix,
            write_mode=write_mode,
            **kwargs
        )
        self._load_table()

    def load_netcdf(
        self,
        in_path: Union[Path, str],
        pattern: str = "**/*.nc",
        field_mapping: dict = None,
        constant_field_values: dict = None,
        location_id_prefix: str = None,
        write_mode: TableWriteEnum = "append",
        **kwargs
    ):
        """Import primary timeseries netcdf data.

        Parameters
        ----------
        in_path : Union[Path, str]
            Path to the timeseries data (file or directory) in
            netcdf file format.
        field_mapping : dict, optional
            A dictionary mapping input fields to output fields.
            Format: {input_field: output_field}
        constant_field_values : dict, optional
            A dictionary mapping field names to constant values.
            Format: {field_name: value}
        location_id_prefix : str, optional
            The prefix to add to location IDs.
            Used to ensure unique location IDs across configurations.
            Note, the methods for fetching USGS and NWM data automatically
            prefix location IDs with "usgs" or the nwm version
            ("nwm12, "nwm21", "nwm22", or "nwm30"), respectively.
        write_mode : TableWriteEnum, optional (default: "append")
            The write mode for the table.
            Options are "append", "upsert", and "overwrite".
            If "append", the table will be appended with new data that does
            already exist.
            If "upsert", existing data will be replaced and new data that
            does not exist will be appended.
            If "overwrite", existing partitions receiving new data are overwritten
        **kwargs
            Additional keyword arguments are passed to xr.open_dataset().

        Includes validation and importing data to database.

        Notes
        -----
        The TEEHR Timeseries table schema includes fields:

        - reference_time
        - value_time
        - configuration_name
        - unit_name
        - variable_name
        - value
        - location_id
        """
        logger.info(f"Loading primary timeseries netcdf data: {in_path}")

        validate_input_is_netcdf(in_path)
        self._load(
            in_path=in_path,
            pattern=pattern,
            field_mapping=field_mapping,
            constant_field_values=constant_field_values,
            location_id_prefix=location_id_prefix,
            write_mode=write_mode,
            **kwargs
        )
        self._load_table()

    def load_fews_xml(
        self,
        in_path: Union[Path, str],
        pattern: str = "**/*.xml",
        field_mapping: dict = {
            "locationId": "location_id",
            "forecastDate": "reference_time",
            "parameterId": "variable_name",
            "units": "unit_name",
            "ensembleId": "configuration_name",
            "ensembleMemberIndex": "member",
            "forecastDate": "reference_time"
        },
        constant_field_values: dict = None,
        location_id_prefix: str = None,
        write_mode: TableWriteEnum = "append",
    ):
        """Import timeseries from XML data format.

        Parameters
        ----------
        in_path : Union[Path, str]
            Path to the timeseries data (file or directory) in
            xml file format.
        pattern : str, optional (default: "**/*.xml")
            The pattern to match files.
        field_mapping : dict, optional
            A dictionary mapping input fields to output fields.
            Format: {input_field: output_field}
            Default mapping:
            {
                "locationId": "location_id",
                "forecastDate": "reference_time",
                "parameterId": "variable_name",
                "units": "unit_name",
                "ensembleId": "configuration_name",
                "ensembleMemberIndex": "member",
                "forecastDate": "reference_time"
            }
        constant_field_values : dict, optional
            A dictionary mapping field names to constant values.
            Format: {field_name: value}.
        location_id_prefix : str, optional
            The prefix to add to location IDs.
            Used to ensure unique location IDs across configurations.
            Note, the methods for fetching USGS and NWM data automatically
            prefix location IDs with "usgs" or the nwm version
            ("nwm12, "nwm21", "nwm22", or "nwm30"), respectively.
        write_mode : TableWriteEnum, optional (default: "append")
            The write mode for the table.
            Options are "append", "upsert", and "overwrite".
            If "append", the table will be appended with new data that does
            already exist.
            If "upsert", existing data will be replaced and new data that
            does not exist will be appended.
            If "overwrite", existing partitions receiving new data are overwritten

        Includes validation and importing data to database.

        Notes
        -----
        This function follows the Delft-FEWS Published Interface (PI)
        XML format.

        reference: https://publicwiki.deltares.nl/display/FEWSDOC/Dynamic+data

        The ``value`` and ``value_time`` fields are parsed automatically.

        The TEEHR Timeseries table schema includes fields:

        - reference_time
        - value_time
        - configuration_name
        - unit_name
        - variable_name
        - value
        - location_id
        - member
        """
        logger.info(f"Loading primary timeseries xml data: {in_path}")

        validate_input_is_xml(in_path)
        self._load(
            in_path=in_path,
            pattern=pattern,
            field_mapping=field_mapping,
            constant_field_values=constant_field_values,
            location_id_prefix=location_id_prefix,
            write_mode=write_mode,
        )
        self._load_table()



"""Dictionaries defining static metric attributes."""
from teehr.models.metrics.basemodels import MetricCategories as mc


ME_ATTRS = {
    "short_name": "ME",
    "display_name": "Mean Error",
    "category": mc.Deterministic,
    "value_range": None,
    "optimal_value": 0.0,
}

RBIAS_ATTRS = {
    "short_name": "RelBias",
    "display_name": "Relative Bias",
    "category": mc.Deterministic,
    "value_range": None,
    "optimal_value": 0.0,
}

MULTBIAS_ATTRS = {
    "short_name": "MultBias",
    "display_name": "Multiplicative Bias",
    "category": mc.Deterministic,
    "value_range": None,
    "optimal_value": 1.0,
}

MSE_ATTRS = {
    "short_name": "MSE",
    "display_name": "Mean Square Error",
    "category": mc.Deterministic,
    "value_range": None,
    "optimal_value": 0.0,
}

RMSE_ATTRS = {
    "short_name": "RMSE",
    "display_name": "Root Mean Square Error",
    "category": mc.Deterministic,
    "value_range": None,
    "optimal_value": 0.0,
}

MAE_ATTRS = {
    "short_name": "MAE",
    "display_name": "Mean Absolute Error",
    "category": mc.Deterministic,
    "value_range": None,
    "optimal_value": 0.0,
}

RMAE_ATTRS = {
    "short_name": "RelMAE",
    "display_name": "Mean Absolute Relative Error",
    "category": mc.Deterministic,
    "value_range": None,
    "optimal_value": 0.0,
}

PEARSON_ATTRS = {
    "short_name": "r",
    "display_name": "Pearson Correlation",
    "category": mc.Deterministic,
    "value_range": [-1.0, 1.0],
    "optimal_value": 1.0,
}

R2_ATTRS = {
    "short_name": "r2",
    "display_name": "Coefficient of Determination",
    "category": mc.Deterministic,
    "value_range": [0.0, 1.0],
    "optimal_value": 1.0,
}

NSE_ATTRS = {
    "short_name": "NSE",
    "display_name": "Nash-Sutcliffe Efficiency",
    "category": mc.Deterministic,
    "value_range": [None, 1.0],
    "optimal_value": 1.0,
}

NNSE_ATTRS = {
    "short_name": "NNSE",
    "display_name": "Normalized Nash-Sutcliffe Efficiency",
    "category": mc.Deterministic,
    "value_range": [0.0, 1.0],
    "optimal_value": 1.0,
}

KGE_ATTRS = {
    "short_name": "KGE",
    "display_name": "Kling-Gupta Efficiency - original",
    "category": mc.Deterministic,
    "value_range": [0.0, 1.0],
    "optimal_value": 1.0,
}

KGE1_ATTRS = {
    "short_name": "KGE_mod1",
    "display_name": "Kling-Gupta Efficiency - modified 1 (2012)",
    "category": mc.Deterministic,
    "value_range": [0.0, 1.0],
    "optimal_value": 1.0,
}

KGE2_ATTRS = {
    "short_name": "KGE_mod2",
    "display_name": "Kling-Gupta Efficiency - modified 2 (2021)",
    "category": mc.Deterministic,
    "value_range": [0.0, 1.0],
    "optimal_value": 1.0,
}

COUNT_ATTRS = {
    "short_name": "count",
    "display_name": "Count",
    "category": mc.Signature,
    "value_range": None,
    "optimal_value": None,
}

MINIMUM_ATTRS = {
    "short_name": "minimum",
    "display_name": "Minimum",
    "category": mc.Signature,
    "value_range": None,
    "optimal_value": None,
}

MAXIMUM_ATTRS = {
    "short_name": "maximum",
    "display_name": "Maximum",
    "category": mc.Signature,
    "value_range": None,
    "optimal_value": None,
}

AVERAGE_ATTRS = {
    "short_name": "average",
    "display_name": "Average",
    "category": mc.Signature,
    "value_range": None,
    "optimal_value": None,
}

SUM_ATTRS = {
    "short_name": "sum",
    "display_name": "Sum",
    "category": mc.Signature,
    "value_range": None,
    "optimal_value": None,
}

VARIANCE_ATTRS = {
    "short_name": "variance",
    "display_name": "Variance",
    "category": mc.Signature,
    "value_range": None,
    "optimal_value": None,
}

MAX_VALUE_DELTA_ATTRS = {
    "short_name": "max_value_delta",
    "display_name": "Max Value Delta",
    "category": mc.Deterministic,
    "value_range": None,
    "optimal_value": None,
}

MAX_VAL_TIME_ATTRS = {
    "short_name": "max_val_time",
    "display_name": "Max Value Time",
    "category": mc.Signature,
    "value_range": None,
    "optimal_value": None,
}

MAX_VALUE_TIMEDELTA_ATTRS = {
    "short_name": "max_value_time_delta",
    "display_name": "Max Value Time Delta",
    "category": mc.Deterministic,
    "value_range": None,
    "optimal_value": None,
    "units": "seconds"
}

ANNUAL_PEAK_RBIAS_ATTRS = {
    "short_name": "annual_peak_relative_bias",
    "display_name": "Annual Peak Relative Bias",
    "category": mc.Deterministic,
    "value_range": None,
    "optimal_value": None,
}

SPEARMAN_R_ATTRS = {
    "short_name": "spearman_correlation",
    "display_name": "Spearman Rank Correlation Coefficient",
    "category": mc.Deterministic,
    "value_range": [-1.0, 1.0],
    "optimal_value": 1.0,
}

RSR_ATTRS = {
    "short_name": "root_mean_standard_deviation_ratio",
    "display_name": "Root Mean Standard Deviation Ratio",
    "category": mc.Deterministic,
    "value_range": [0.0, None],
    "optimal_value": 0.0,
}

CRPS_ENSEMBLE_ATTRS = {
    "short_name": "crps_ensemble",
    "display_name": "Continuous Ranked Probability Score - Ensemble",
    "category": mc.Probabilistic,
    "value_range": [None, None],
    "optimal_value": 0.0,
}



"""Location table class."""
import teehr.const as const
from teehr.evaluation.tables.base_table import BaseTable
from teehr.loading.locations import convert_locations
from teehr.models.filters import LocationFilter
from teehr.models.table_enums import LocationFields
from teehr.querying.utils import df_to_gdf
import teehr.models.pandera_dataframe_schemas as schemas
from pathlib import Path
from typing import Union
import logging
from teehr.utils.utils import to_path_or_s3path, remove_dir_if_exists
from teehr.loading.utils import add_or_replace_sdf_column_prefix
from teehr.models.table_enums import TableWriteEnum


logger = logging.getLogger(__name__)


class LocationTable(BaseTable):
    """Access methods to locations table."""

    def __init__(self, ev):
        """Initialize class."""
        super().__init__(ev)
        self.name = "locations"
        self.dir = to_path_or_s3path(ev.dataset_dir, self.name)
        self.format = "parquet"
        self.filter_model = LocationFilter
        self.schema_func = schemas.locations_schema
        self.unique_column_set = ["id"]

    def field_enum(self) -> LocationFields:
        """Get the location fields enum."""
        fields = self._get_schema("pandas").columns.keys()
        return LocationFields(
            "LocationFields",
            {field: field for field in fields}
        )

    def to_pandas(self):
        """Return Pandas DataFrame for Location Table."""
        self._check_load_table()
        df = self.df.toPandas()
        df.attrs['table_type'] = 'locations'
        df.attrs['fields'] = self.fields()
        return df

    def to_geopandas(self):
        """Return GeoPandas DataFrame."""
        self._check_load_table()
        gdf = df_to_gdf(self.to_pandas())
        gdf.attrs['table_type'] = self.name
        gdf.attrs['fields'] = self.fields()
        return gdf

    def load_spatial(
        self,
        in_path: Union[Path, str],
        field_mapping: dict = None,
        pattern: str = "**/*.parquet",
        location_id_prefix: str = None,
        write_mode: TableWriteEnum = "append",
        **kwargs
    ):
        """Import geometry data.

        Parameters
        ----------
        in_path : Union[Path, str]
            The input file or directory path.
            Any file format that can be read by GeoPandas.
        field_mapping : dict, optional
            A dictionary mapping input fields to output fields.
            Format: {input_field: output_field}
        pattern : str, optional (default: "**/*.parquet")
            The pattern to match files.
            Only used when in_path is a directory.
        location_id_prefix : str, optional
            The prefix to add to location IDs.
            Used to ensure unique location IDs across configurations.
            Note, the methods for fetching USGS and NWM data automatically
            prefix location IDs with "usgs" or the nwm version
            ("nwm12, "nwm21", "nwm22", or "nwm30"), respectively.
        write_mode : TableWriteEnum, optional (default: "append")
            The write mode for the table.
            Options are "append", "upsert", and "overwrite".
            If "append", the table will be appended with new data that does
            already exist.
            If "upsert", existing data will be replaced and new data that
            does not exist will be appended.
            If "overwrite", existing partitions receiving new data are
            overwritten.
        **kwargs
            Additional keyword arguments are passed to GeoPandas read_file().

        File is first read by GeoPandas, field names renamed and
        then validated and inserted into the dataset.

        Notes
        -----
        The TEEHR Location Crosswalk table schema includes fields:

        - id
        - name
        - geometry

        ..note::
          The methods for fetching USGS and NWM data expect
          location IDs to be prefixed with "usgs" or the nwm version
          ("nwm12, "nwm21", "nwm22", or "nwm30"), respectively.
        """
        cache_dir = Path(
            self.ev.dir_path,
            const.CACHE_DIR,
            const.LOADING_CACHE_DIR,
            const.LOCATIONS_DIR
        )
        # Clear the cache directory if it exists.
        remove_dir_if_exists(cache_dir)

        convert_locations(
            in_path,
            cache_dir,
            field_mapping=field_mapping,
            pattern=pattern,
            **kwargs
        )
        # Read the converted files to Spark DataFrame
        df = self._read_files(cache_dir)

        # Add or replace location_id prefix if provided
        if location_id_prefix:
            df = add_or_replace_sdf_column_prefix(
                sdf=df,
                column_name="id",
                prefix=location_id_prefix,
            )

        # Validate using the validate method
        validated_df = self._validate(df)

        # Write to the table
        self._write_spark_df(
            df=validated_df,
            write_mode=write_mode,
            num_partitions=1,
        )

        # Reload the table
        self._load_table()


"""Data types for the TEEHR schema."""
from datetime import datetime

import numpy as np

VALUE = "value"
VALUE_TIME = "value_time"
REFERENCE_TIME = "reference_time"
LOCATION_ID = "location_id"
UNIT_NAME = "unit_name"
VARIABLE_NAME = "variable_name"
CONFIGURATION_NAME = "configuration_name"
MEMBER = "member"

USGS_NODATA_VALUES = [-999999, -999, -9999, -99999]
USGS_CONFIGURATION_NAME = "usgs_observations"

NWM_BUCKET = "national-water-model"
NWM22_UNIT_LOOKUP = {"m3 s-1": "m^3/s"}

NWM30_START_DATE = datetime(2023, 9, 19, 0)
NWM21_START_DATE = datetime(2021, 4, 20, 0)  # v2.1 and 2.2 are the same
NWM20_START_DATE = datetime(2019, 6, 19, 0)
NWM12_START_DATE = datetime(2018, 9, 17, 0)

NWM_S3_JSON_PATH = "s3://ciroh-nwm-zarr-copy"

USGS_VARIABLE_MAPPER = {
    VARIABLE_NAME: {
        "iv": "streamflow_hourly_inst",
        "dv": "streamflow_daily_mean",
    },
    UNIT_NAME: {
        "SI": "m^3/s",
        "Imperial": "ft^3/s",
    },
}

NWM_VARIABLE_MAPPER = {
    VARIABLE_NAME: {
        "streamflow": "streamflow_hourly_inst",
        "RAINRATE": "rainfall_hourly_rate",
    },
    UNIT_NAME: {
        "m3 s-1": "m^3/s",
        "mm s^-1": "mm/s",  # NWM 3.0 forcing
        "mm s-1": "mm/s",   # NWM 2.2 forcing
    },
}

NWM12_ANALYSIS_CONFIG = {
    "analysis_assim": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim",
    },
    "forcing_analysis_assim": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim",
    }
}

NWM20_ANALYSIS_CONFIG = {
    "analysis_assim": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim",
    },
    "analysis_assim_extend": {
        "num_lookback_hrs": 28,
        "cycle_z_hours": [16],
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim_extend",
    },
    "analysis_assim_long": {
        "num_lookback_hrs": 12,
        "cycle_z_hours": np.arange(0, 24, 6),
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim_long",
    },
    "analysis_assim_hawaii": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "hawaii",
        "configuration_name_in_filepath": "analysis_assim",
    },
    "forcing_analysis_assim": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim",
    },
    "forcing_analysis_assim_extend": {
        "num_lookback_hrs": 28,
        "cycle_z_hours": [16],
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim_extend",
    },
    "forcing_analysis_assim_hawaii": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "hawaii",
        "configuration_name_in_filepath": "analysis_assim",
    }
}


NWM22_ANALYSIS_CONFIG = {
    "analysis_assim": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim",
    },
    "analysis_assim_no_da": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim_no_da",
    },
    "analysis_assim_extend": {
        "num_lookback_hrs": 28,
        "cycle_z_hours": [16],
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim_extend",
    },
    "analysis_assim_extend_no_da": {
        "num_lookback_hrs": 28,
        "cycle_z_hours": [16],
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim_extend_no_da",
    },
    "analysis_assim_long": {
        "num_lookback_hrs": 12,
        "cycle_z_hours": np.arange(0, 24, 6),
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim_long",
    },
    "analysis_assim_long_no_da": {
        "num_lookback_hrs": 12,
        "cycle_z_hours": np.arange(0, 24, 6),
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim_long_no_da",
    },
    "analysis_assim_hawaii": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "hawaii",
        "configuration_name_in_filepath": "analysis_assim",
    },
    "analysis_assim_hawaii_no_da": {
        "num_lookback_hrs": 3,
        "cycle_cycle_z_hourstimes": np.arange(0, 24, 1),
        "domain": "hawaii",
        "configuration_name_in_filepath": "analysis_assim_no_da",
    },
    "analysis_assim_puertorico": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "puertorico",
        "configuration_name_in_filepath": "analysis_assim",
    },
    "analysis_assim_puertorico_no_da": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "puertorico",
        "configuration_name_in_filepath": "analysis_assim_no_da",
    },
    "forcing_analysis_assim": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim",
    },
    "forcing_analysis_assim_extend": {
        "num_lookback_hrs": 28,
        "cycle_z_hours": [16],
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim_extend",
    },
    "forcing_analysis_assim_hawaii": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "hawaii",
        "configuration_name_in_filepath": "analysis_assim",
    },
    "forcing_analysis_assim_puertorico": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "puertorico",
        "configuration_name_in_filepath": "analysis_assim",
    },
}

NWM30_ANALYSIS_CONFIG = {
    "analysis_assim": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim",
    },
    "analysis_assim_no_da": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim_no_da",
    },
    "analysis_assim_extend": {
        "num_lookback_hrs": 28,
        "cycle_z_hours": [16],
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim_extend",
    },
    "analysis_assim_extend_no_da": {
        "num_lookback_hrs": 28,
        "cycle_z_hours": [16],
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim_extend_no_da",
    },
    "analysis_assim_long": {
        "num_lookback_hrs": 12,
        "cycle_z_hours": np.arange(0, 24, 6),
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim_long",
    },
    "analysis_assim_long_no_da": {
        "num_lookback_hrs": 12,
        "cycle_z_hours": np.arange(0, 24, 6),
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim_long_no_da",
    },
    "analysis_assim_alaska": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "alaska",
        "configuration_name_in_filepath": "analysis_assim",
    },
    "analysis_assim_alaska_no_da": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "alaska",
        "configuration_name_in_filepath": "analysis_assim_no_da",
    },
    "analysis_assim_extend_alaska": {
        "num_lookback_hrs": 32,
        "cycle_z_hours": [20],
        "domain": "alaska",
        "configuration_name_in_filepath": "analysis_assim_extend",
    },
    "analysis_assim_extend_alaska_no_da": {
        "num_lookback_hrs": 32,
        "cycle_z_hours": [20],
        "domain": "alaska",
        "configuration_name_in_filepath": "analysis_assim_extend_no_da",
    },
    "analysis_assim_hawaii": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "hawaii",
        "configuration_name_in_filepath": "analysis_assim",
    },
    "analysis_assim_hawaii_no_da": {
        "num_lookback_hrs": 3,
        "cycle_cycle_z_hourstimes": np.arange(0, 24, 1),
        "domain": "hawaii",
        "configuration_name_in_filepath": "analysis_assim_no_da",
    },
    "analysis_assim_puertorico": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "puertorico",
        "configuration_name_in_filepath": "analysis_assim",
    },
    "analysis_assim_puertorico_no_da": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "puertorico",
        "configuration_name_in_filepath": "analysis_assim_no_da",
    },
    "forcing_analysis_assim": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim",
    },
    "forcing_analysis_assim_extend": {
        "num_lookback_hrs": 28,
        "cycle_z_hours": [16],
        "domain": "conus",
        "configuration_name_in_filepath": "analysis_assim_extend",
    },
    "forcing_analysis_assim_alaska": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "alaska",
        "configuration_name_in_filepath": "analysis_assim",
    },
    "forcing_analysis_assim_hawaii": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "hawaii",
        "configuration_name_in_filepath": "analysis_assim",
    },
    "forcing_analysis_assim_puertorico": {
        "num_lookback_hrs": 3,
        "cycle_z_hours": np.arange(0, 24, 1),
        "domain": "puertorico",
        "configuration_name_in_filepath": "analysis_assim",
    },
}


# WKT strings extracted from NWM grids
CONUS_NWM_WKT = 'PROJCS["Lambert_Conformal_Conic",GEOGCS["GCS_Sphere",DATUM["D_Sphere",SPHEROID["Sphere",6370000.0,0.0]], \
PRIMEM["Greenwich",0.0],UNIT["Degree",0.0174532925199433]],PROJECTION["Lambert_Conformal_Conic_2SP"],PARAMETER["false_easting",0.0],\
PARAMETER["false_northing",0.0],PARAMETER["central_meridian",-97.0],PARAMETER["standard_parallel_1",30.0],\
PARAMETER["standard_parallel_2",60.0],PARAMETER["latitude_of_origin",40.0],UNIT["Meter",1.0]]' # noqa

HI_NWM_WKT = 'PROJCS["Lambert_Conformal_Conic",GEOGCS["GCS_Sphere",DATUM["D_Sphere",SPHEROID["Sphere",6370000.0,0.0]],\
PRIMEM["Greenwich",0.0],UNIT["Degree",0.0174532925199433]],PROJECTION["Lambert_Conformal_Conic_2SP"],PARAMETER["false_easting",0.0],\
PARAMETER["false_northing",0.0],PARAMETER["central_meridian",-157.42],PARAMETER["standard_parallel_1",10.0],\
PARAMETER["standard_parallel_2",30.0],PARAMETER["latitude_of_origin",20.6],UNIT["Meter",1.0]]' # noqa

PR_NWM_WKT = 'PROJCS["Sphere_Lambert_Conformal_Conic",GEOGCS["GCS_Sphere",DATUM["D_Sphere",SPHEROID["Sphere",6370000.0,0.0]],\
PRIMEM["Greenwich",0.0],UNIT["Degree",0.0174532925199433]],PROJECTION["Lambert_Conformal_Conic_2SP"],PARAMETER["false_easting",0.0],\
PARAMETER["false_northing",0.0],PARAMETER["central_meridian",-65.91],PARAMETER["standard_parallel_1",18.1],\
PARAMETER["standard_parallel_2",18.1],PARAMETER["latitude_of_origin",18.1],UNIT["Meter",1.0]]' # noqa

AL_NWM_WKT = 'PROJCS["Sphere_Stereographic",GEOGCS["Sphere",DATUM["Sphere",SPHEROID["unnamed",6370000,0]], \
PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]]], \
PROJECTION["Polar_Stereographic"],PARAMETER["latitude_of_origin",60],PARAMETER["central_meridian",-135], \
PARAMETER["false_easting",0],PARAMETER["false_northing",0],UNIT["metre",1],AXIS["Easting",SOUTH], \
AXIS["Northing",SOUTH]]' # noqa


"""Module contains S3Path.
A class that represents an S3 path and allows for easy concatenation of paths.
"""
from typing import Tuple

class S3Path():
    """Class that represents an S3 path and allows for easy concatenation of paths."""

    def __init__(self, path: str, *args: Tuple[str, ...]):
        self.path = str(path)

        # check that the path starts with "s3://", "s3a://" or "s3n://"
        if (
            not self.path.startswith("s3://") and
            not self.path.startswith("s3a://") and
            not self.path.startswith("s3n://")
        ):
            raise ValueError("Invalid S3 path")

        if args:
            for arg in args:
                if not isinstance(arg, str):
                    raise ValueError("Only strings can be concatenated to S3Path")
                self.path = self.path + "/" + arg

    def __truediv__(self, other):
        if not isinstance(other, str):
            raise ValueError("Only strings can be concatenated to S3Path")
        return S3Path(self.path + "/" + other)

    # string representation of the path
    def __str__(self):
        return str(self.path)

    def __repr__(self):
        return f"S3Path({self.path})"


"""Functions for probabilistic metric calculations in Spark queries."""
from typing import Dict, Callable
import logging

import pandas as pd
import numpy as np
import scoringrules as sr

from teehr.models.metrics.basemodels import MetricsBasemodel
from teehr.metrics.deterministic_funcs import _transform

logger = logging.getLogger(__name__)


def _pivot_by_value_time(
    p: pd.Series,
    s: pd.Series,
    value_time: pd.Series
) -> Dict:
    """Pivot the timeseries data by value_time.

    Notes
    -----
    prim_arr: A 1-D array of observations at each time step in the forecast.

    sec_arr: A 2-D array of simulations at each time step in the forecast.
            The first dimension should be the time step, and the second
            dimension should be the ensemble member.
    """
    # TODO: Probably a better way to do this?
    primary = []
    secondary = []
    for vt in value_time.unique():
        vt_index = value_time[value_time == vt].index
        primary.append(p[vt_index].values[0])
        secondary.append(s[vt_index].values)
    return {"primary": np.array(primary), "secondary": np.array(secondary)}


def create_crps_func(model: MetricsBasemodel) -> Callable:
    """Create the CRPS ensemble metric function."""
    logger.debug("Building the CRPS ensemble metric func.")

    def ensemble_crps(
        p: pd.Series,
        s: pd.Series,
        value_time: pd.Series,
    ) -> float:
        """Create a wrapper around scoringrules crps_ensemble.

        Parameters
        ----------
        p : pd.Series
            The primary values.
        s : pd.Series
            The secondary values.
        value_time : pd.Series
            The value time.

        Returns
        -------
        float
            The mean Continuous Ranked Probability Score (CRPS) for the
            ensemble, either as a single value or array of values.
        """
        p, s, value_time = _transform(p, s, model, value_time)
        pivoted_dict = _pivot_by_value_time(p, s, value_time)

        if model.summary_func is not None:
            return model.summary_func(
                sr.crps_ensemble(
                    pivoted_dict["primary"],
                    pivoted_dict["secondary"],
                    estimator=model.estimator,
                    backend=model.backend
                )
            )
        else:
            return sr.crps_ensemble(
                pivoted_dict["primary"],
                pivoted_dict["secondary"],
                estimator=model.estimator,
                backend=model.backend
            )

    return ensemble_crps



"""Evaluation module."""
from datetime import datetime
from typing import Union, Literal, List
from pathlib import Path
from teehr.evaluation.tables.attribute_table import AttributeTable
from teehr.evaluation.tables.configuration_table import ConfigurationTable
from teehr.evaluation.tables.location_attribute_table import LocationAttributeTable
from teehr.evaluation.tables.location_crosswalk_table import LocationCrosswalkTable
from teehr.evaluation.tables.location_table import LocationTable
from teehr.evaluation.tables.primary_timeseries_table import PrimaryTimeseriesTable
from teehr.evaluation.tables.secondary_timeseries_table import SecondaryTimeseriesTable
from teehr.evaluation.tables.unit_table import UnitTable
from teehr.evaluation.tables.variable_table import VariableTable
from teehr.evaluation.tables.joined_timeseries_table import JoinedTimeseriesTable
from teehr.utils.s3path import S3Path
from teehr.utils.utils import to_path_or_s3path, remove_dir_if_exists
from pyspark.sql import SparkSession
from pyspark import SparkConf
import logging
from teehr.loading.utils import copy_template_to
from teehr.loading.s3.clone_from_s3 import (
    list_s3_evaluations,
    clone_from_s3
)
import teehr.const as const
from teehr.evaluation.fetch import Fetch
from teehr.evaluation.metrics import Metrics
import pandas as pd
from teehr.visualization.dataframe_accessor import TEEHRDataFrameAccessor # noqa


logger = logging.getLogger(__name__)


class Evaluation:
    """The Evaluation class.

    This is the main class for the TEEHR evaluation.
    """

    def __init__(
        self,
        dir_path: Union[str, Path, S3Path],
        create_dir: bool = False,
        spark: SparkSession = None
    ):
        """
        Initialize the Evaluation class.

        Parameters
        ----------
        dir_path : Union[str, Path, S3Path]
            The path to the evaluation directory.
        spark : SparkSession, optional
            The SparkSession object, by default None
        """
        self.dir_path = to_path_or_s3path(dir_path)

        self.is_s3 = False
        if isinstance(self.dir_path, S3Path):
            self.is_s3 = True
            logger.info(f"Using S3 path {self.dir_path}.  Evaluation will be read-only")

        self.spark = spark

        self.dataset_dir = to_path_or_s3path(
            self.dir_path, const.DATASET_DIR
        )
        self.cache_dir = to_path_or_s3path(
            self.dir_path, const.CACHE_DIR
        )
        self.scripts_dir = to_path_or_s3path(
            self.dir_path, const.SCRIPTS_DIR
        )

        if not self.is_s3 and not Path(self.dir_path).is_dir():
            if create_dir:
                logger.info(f"Creating directory {self.dir_path}.")
                Path(self.dir_path).mkdir(parents=True, exist_ok=True)
            else:
                logger.error(f"Directory {self.dir_path} does not exist.")
                raise NotADirectoryError

        # Create a local Spark Session if one is not provided.
        if not self.spark:
            logger.info("Creating a new Spark session.")
            conf = (
                SparkConf()
                .setAppName("TEEHR")
                .setMaster("local[*]")
                .set("spark.sql.sources.partitionOverwriteMode", "dynamic")
                .set("spark.hadoop.fs.s3a.impl", "org.apache.hadoop.fs.s3a.S3AFileSystem")
                .set("spark.hadoop.fs.s3a.aws.credentials.provider", "org.apache.hadoop.fs.s3a.AnonymousAWSCredentialsProvider")
                .set("spark.sql.execution.arrow.pyspark.enabled", "true")
                .set("spark.sql.session.timeZone", "UTC")
            )
            self.spark = SparkSession.builder.config(conf=conf).getOrCreate()

    @property
    def fetch(self) -> Fetch:
        """The fetch component class for accessing external data."""
        if self.is_s3:
            logger.error("Cannot fetch data and save to S3 yet.")
            raise Exception("Cannot fetch data and save to S3 yet.")

        return Fetch(self)

    @property
    def metrics(self) -> Metrics:
        """The metrics component class for calculating performance metrics."""
        return Metrics(self)

    @property
    def units(self) -> UnitTable:
        """Access the units table."""
        return UnitTable(self)

    @property
    def variables(self) -> VariableTable:
        """Access the variables table."""
        return VariableTable(self)

    @property
    def attributes(self) -> AttributeTable:
        """Access the attributes table."""
        return AttributeTable(self)

    @property
    def configurations(self) -> ConfigurationTable:
        """Access the configurations table."""
        return ConfigurationTable(self)

    @property
    def locations(self) -> LocationTable:
        """Access the locations table."""
        return LocationTable(self)

    @property
    def location_attributes(self) -> LocationAttributeTable:
        """Access the location attributes table."""
        return LocationAttributeTable(self)

    @property
    def location_crosswalks(self) -> LocationCrosswalkTable:
        """Access the location crosswalks table."""
        return LocationCrosswalkTable(self)

    @property
    def primary_timeseries(self) -> PrimaryTimeseriesTable:
        """Access the primary timeseries table."""
        return PrimaryTimeseriesTable(self)

    @property
    def secondary_timeseries(self) -> SecondaryTimeseriesTable:
        """Access the secondary timeseries table."""
        return SecondaryTimeseriesTable(self)

    @property
    def joined_timeseries(self) -> JoinedTimeseriesTable:
        """Access the joined timeseries table."""
        return JoinedTimeseriesTable(self)

    def enable_logging(self):
        """Enable logging."""
        logger = logging.getLogger("teehr")
        # logger.addHandler(logging.StreamHandler())
        if self.is_s3:
            logger_path = Path(Path.home, 'teehr.log')
        else:
            logger_path = Path(self.dir_path, 'teehr.log')

        handler = logging.FileHandler(logger_path)
        handler.setFormatter(
            logging.Formatter(
                "%(asctime)s %(levelname)s %(message)s"
            )
        )
        logger.addHandler(
            handler
        )
        logger.setLevel(logging.DEBUG)

    def clone_template(self):
        """Create a study from the standard template.

        This method mainly copies the template directory to the specified
        evaluation directory.
        """
        if self.is_s3:
            logger.error("Cannot clone template to S3.")
            raise Exception("Cannot clone template to S3.")

        teehr_root = Path(__file__).parent.parent
        template_dir = Path(teehr_root, "template")
        logger.info(f"Copying template from {template_dir} to {self.dir_path}")
        copy_template_to(template_dir, self.dir_path)

    @staticmethod
    def list_s3_evaluations(
        format: Literal["pandas", "list"] = "pandas"
    ) -> Union[list, pd.DataFrame]:
        """List the evaluations available on S3.

        Parameters
        ----------
        format : str, optional
            The format of the output. Either "pandas" or "list".
            The default is "pandas".
        """
        return list_s3_evaluations(format=format)

    def clone_from_s3(
        self,
        evaluation_name: str,
        primary_location_ids: List[str] = None,
        start_date: Union[str, datetime] = None,
        end_date: Union[str, datetime] = None,
    ):
        """Fetch the study data from S3.

        Copies the study from s3 to the local directory, with the option
        to subset the dataset by primary location ID, start and end dates.

        Parameters
        ----------
        evaluation_name : str
            The name of the evaluation to clone from S3.
            Use the list_s3_evaluations method to get the available
            evaluations.
        primary_location_ids : List[str], optional
            The list of primary location ids to subset the data.
            The default is None.
        start_date : Union[str, datetime], optional
            The start date to subset the data.
            The default is None.
        end_date : Union[str, datetime], optional
            The end date to subset the data.
            The default is None.

        Notes
        -----

        Includes the following tables:
            - units
            - variables
            - attributes
            - configurations
            - locations
            - location_attributes
            - location_crosswalks
            - primary_timeseries
            - secondary_timeseries
            - joined_timeseries

        Also includes the user_defined_fields.py script.

        """
        if self.is_s3:
            logger.error("Cannot clone from S3 to S3.")
            raise Exception("Cannot clone from S3 to S3.")

        return clone_from_s3(
            self,
            evaluation_name,
            primary_location_ids,
            start_date,
            end_date
        )

    def clean_cache(self):
        """Clean temporary files.

        Includes removing temporary files.
        """
        if self.is_s3:
            logger.error("Cannot clean cache on S3.")
            raise Exception("Cannot clean cache on S3.")

        logger.info(f"Removing temporary files from {self.cache_dir}")
        remove_dir_if_exists(self.cache_dir)
        self.cache_dir.mkdir()

    def sql(self, query: str, create_temp_views: List[str] = None):
        """Run a SQL query on the Spark session against the TEEHR tables.

        Parameters
        ----------
        query : str
            The SQL query to run.
        create_temp_views : List[str], optional
            A list of tables to create temporary views for.
            The default is None which creates all.

        Returns
        -------
        pyspark.sql.DataFrame
            The result of the SQL query.
            This is lazily evaluated so you need to call an action (e.g., sdf.show()) to get the result.

        By default this method has access to the following tables preloaded as temporary views:
            - units
            - variables
            - attributes
            - configurations
            - locations
            - location_attributes
            - location_crosswalks
            - primary_timeseries
            - secondary_timeseries
            - joined_timeseries
        """
        if not create_temp_views:
            create_temp_views = [
                "units",
                "variables",
                "attributes",
                "configurations",
                "locations",
                "location_attributes",
                "location_crosswalks",
                "primary_timeseries",
                "secondary_timeseries",
                "joined_timeseries"
            ]

        if "units" in create_temp_views:
            self.units.to_sdf().createOrReplaceTempView("units")
        if "variables" in create_temp_views:
            self.variables.to_sdf().createOrReplaceTempView("variables")
        if "attributes" in create_temp_views:
            self.attributes.to_sdf().createOrReplaceTempView("attributes")
        if "configurations" in create_temp_views:
            self.configurations.to_sdf().createOrReplaceTempView("configurations")
        if "locations" in create_temp_views:
            self.locations.to_sdf().createOrReplaceTempView("locations")
        if "location_attributes" in create_temp_views:
            self.location_attributes.to_sdf().createOrReplaceTempView("location_attributes")
        if "location_crosswalks" in create_temp_views:
            self.location_crosswalks.to_sdf().createOrReplaceTempView("location_crosswalks")
        if "primary_timeseries" in create_temp_views:
            self.primary_timeseries.to_sdf().createOrReplaceTempView("primary_timeseries")
        if "secondary_timeseries" in create_temp_views:
            self.secondary_timeseries.to_sdf().createOrReplaceTempView("secondary_timeseries")
        if "joined_timeseries" in create_temp_views:
            self.joined_timeseries.to_sdf().createOrReplaceTempView("joined_timeseries")

        return self.spark.sql(query)



"""Signature functions."""
import pandas as pd
import numpy as np

from teehr.models.metrics.basemodels import MetricsBasemodel
from teehr.models.metrics.basemodels import TransformEnum
from typing import Callable, Optional
import logging
logger = logging.getLogger(__name__)


def _transform(
        p: pd.Series,
        model: MetricsBasemodel,
        t: Optional[pd.Series] = None
) -> tuple:
    """Apply timeseries transform for metrics calculations."""
    # Apply transform
    if model.transform is not None:
        match model.transform:
            case TransformEnum.log:
                logger.debug("Applying log transform")
                p = np.log(p)
            case TransformEnum.sqrt:
                logger.debug("Applying square root transform")
                p = np.sqrt(p)
            case TransformEnum.square:
                logger.debug("Applying square transform")
                p = np.square(p)
            case TransformEnum.cube:
                logger.debug("Applying cube transform")
                p = np.power(p, 3)
            case TransformEnum.exp:
                logger.debug("Applying exponential transform")
                p = np.exp(p)
            case TransformEnum.inv:
                logger.debug("Applying inverse transform")
                p = 1.0 / p
            case TransformEnum.abs:
                logger.debug("Applying absolute value transform")
                p = np.abs(p)
            case _:
                raise ValueError(
                    f"Unsupported transform: {model.transform}"
                )
    else:
        logger.debug("No transform specified, using original values")

    # Remove invalid values and align series
    if t is not None:
        if isinstance(t, pd.Series):
            valid_mask = np.isfinite(p)
            p = p[valid_mask]
            t = t[valid_mask]
        else:
            raise TypeError(
                f"Invalid type for t: {type(t)}. Expected pd.Series."
            )
    else:
        valid_mask = np.isfinite(p)
        p = p[valid_mask]

    if t is not None:
        return p, t
    else:
        return p


def mvt_wrapper(model: MetricsBasemodel) -> Callable:
    """Create max_value_time metric function."""
    logger.debug("Building the max_value_time metric function")

    def max_value_time(
        p: pd.Series,
        value_time: pd.Series
    ) -> pd.Timestamp:
        """Max value time."""
        p, value_time = _transform(p, model, value_time)
        return value_time[p.idxmax()]

    return max_value_time


def variance_wrapper(model: MetricsBasemodel) -> Callable:
    """Create variance metric function."""
    logger.debug("Building the variance metric function")

    def variance(p: pd.Series) -> float:
        """Variance."""
        p = _transform(p, model)
        return np.var(p)

    return variance


def count_wrapper(model: MetricsBasemodel) -> Callable:
    """Create count metric function."""
    logger.debug("Building the count metric function")

    def count(p: pd.Series) -> float:
        """Count."""
        p = _transform(p, model)
        return len(p)

    return count


def min_wrapper(model: MetricsBasemodel) -> Callable:
    """Create minimum metric function."""
    logger.debug("Building the minimum metric function")

    def minimum(p: pd.Series) -> float:
        """Minimum."""
        p = _transform(p, model)
        return np.min(p)

    return minimum


def max_wrapper(model: MetricsBasemodel) -> Callable:
    """Create maximum metric function."""
    logger.debug("Building the maximum metric function")

    def maximum(p: pd.Series) -> float:
        """Maximum."""
        p = _transform(p, model)
        return np.max(p)

    return maximum


def avg_wrapper(model: MetricsBasemodel) -> Callable:
    """Create average metric function."""
    logger.debug("Building the average metric function")

    def average(p: pd.Series) -> float:
        """Average."""
        p = _transform(p, model)
        return np.mean(p)

    return average


def sum_wrapper(model: MetricsBasemodel) -> Callable:
    """Create sum metric function."""
    logger.debug("Building the sum metric function")

    def sum(p: pd.Series) -> float:
        """Sum."""
        p = _transform(p, model)
        return np.sum(p)

    return sum



"""A utility for rewriting Evaluations with reference time as a partition."""
from pathlib import Path
from typing import List
import pyspark.sql as ps
import shutil

import teehr
from teehr.utils.utils import remove_dir_if_exists

PARTITION_BY = [
    "configuration_name",
    "variable_name",
    "reference_time"
]
KWARGS = {
    "header": "true",
}


def _check_for_null_reference_time(partition_by: List[str], sdf: ps.DataFrame):
    """Check if the reference time column is all null."""
    if "reference_time" in sdf.columns:
        if len(sdf.filter(sdf.reference_time.isNotNull()).collect()) == 0:
            if "reference_time" in partition_by:
                partition_by.remove("reference_time")
    return partition_by


def rewrite_secondary_timeseries(
    ev: teehr.Evaluation,
    partition_by: List[str] = PARTITION_BY
):
    """Rewrite secondary timeseries partitioning on reference time."""
    secondary_sdf = ev.secondary_timeseries.to_sdf()

    partition_by = _check_for_null_reference_time(
        partition_by,
        secondary_sdf
    )
    temp_path = Path(
        ev.cache_dir,
        "temp_rewrite",
        "secondary_timeseries"
    )
    (
        secondary_sdf.
        write.
        partitionBy(partition_by).
        format("parquet").
        mode("overwrite").
        options(**KWARGS).
        save(str(temp_path))
    )
    # Copy in temp dir then remove
    remove_dir_if_exists(ev.secondary_timeseries.dir)
    shutil.copytree(
        str(temp_path),
        str(ev.secondary_timeseries.dir),
        dirs_exist_ok=True
    )
    remove_dir_if_exists(Path(ev.cache_dir, "temp_rewrite"))


def rewrite_primary_timeseries(
    ev: teehr.Evaluation,
    partition_by: List[str] = PARTITION_BY
):
    """Rewrite primary timeseries partitioning on reference time."""
    primary_sdf = ev.primary_timeseries.to_sdf()

    partition_by = _check_for_null_reference_time(
        partition_by,
        primary_sdf
    )
    temp_path = Path(
        ev.cache_dir,
        "temp_rewrite",
        "primary_timeseries"
    )
    (
        primary_sdf.
        write.
        partitionBy(partition_by).
        format("parquet").
        mode("overwrite").
        options(**KWARGS).
        save(str(temp_path))
    )
    # Copy in temp dir then remove
    remove_dir_if_exists(ev.primary_timeseries.dir)
    shutil.copytree(
        str(temp_path),
        str(ev.primary_timeseries.dir),
        dirs_exist_ok=True
    )
    remove_dir_if_exists(Path(ev.cache_dir, "temp_rewrite"))


def rewrite_joined_timeseries(
    ev: teehr.Evaluation,
    partition_by: List[str] = PARTITION_BY
):
    """Rewrite joined timeseries partitioning on reference time."""
    joined_sdf = ev.joined_timeseries.to_sdf()

    partition_by = _check_for_null_reference_time(
        partition_by,
        joined_sdf
    )
    temp_path = Path(
        ev.cache_dir,
        "temp_rewrite",
        "joined_timeseries"
    )
    (
        joined_sdf.
        write.
        partitionBy(partition_by).
        format("parquet").
        mode("overwrite").
        options(**KWARGS).
        save(str(temp_path))
    )
    # Copy in temp dir then remove
    remove_dir_if_exists(ev.joined_timeseries.dir)
    shutil.copytree(
        str(temp_path),
        str(ev.joined_timeseries.dir),
        dirs_exist_ok=True
    )
    remove_dir_if_exists(Path(ev.cache_dir, "temp_rewrite"))


if __name__ == "__main__":

    # Define the path to your evaluation directory
    test_eval_dir = Path("/mnt/data/ciroh/teehr/test_stuff/temp_repartition_ev/temp_ev")

    # Create an Evaluation object and create the directory
    ev = teehr.Evaluation(dir_path=test_eval_dir)

    # Re-write the primary timeseries with reference time as a partition,
    # if it is not null.
    rewrite_primary_timeseries(ev)

    # Re-write the secondary timeseries with reference time as a partition,
    # if it is not null.
    rewrite_secondary_timeseries(ev)

    # Re-write the joined timeseries with reference time as a partition,
    # if it is not null.
    rewrite_joined_timeseries(ev)

    pass



"""Component class for fetching data from external sources."""
from typing import Union, List, Optional
from datetime import datetime
import logging
from pathlib import Path

import pandas as pd

from teehr.utils.utils import remove_dir_if_exists
import teehr.const as const
from teehr.fetching.usgs.usgs import usgs_to_parquet
from teehr.fetching.nwm.nwm_points import nwm_to_parquet
from teehr.fetching.nwm.nwm_grids import nwm_grids_to_parquet
from teehr.fetching.nwm.retrospective_points import nwm_retro_to_parquet
from teehr.fetching.nwm.retrospective_grids import nwm_retro_grids_to_parquet
from teehr.loading.timeseries import (
    validate_and_insert_timeseries,
)
from teehr.fetching.utils import (
    format_nwm_variable_name,
    format_nwm_configuration_name
)
from teehr.models.fetching.nwm22_grid import ForcingVariablesEnum
from teehr.models.fetching.utils import (
    USGSChunkByEnum,
    USGSServiceEnum,
    SupportedNWMRetroVersionsEnum,
    SupportedNWMRetroDomainsEnum,
    NWMChunkByEnum,
    ChannelRtRetroVariableEnum,
    SupportedNWMOperationalVersionsEnum,
    SupportedNWMDataSourcesEnum,
    SupportedKerchunkMethod,
    TimeseriesTypeEnum
)
from teehr.models.table_enums import TableWriteEnum
from teehr.fetching.const import (
    USGS_CONFIGURATION_NAME,
    USGS_VARIABLE_MAPPER,
    VARIABLE_NAME,
    NWM_VARIABLE_MAPPER
)
from teehr.models.pydantic_table_models import (
    Configuration
)

logger = logging.getLogger(__name__)


class Fetch:
    """Component class for fetching data from external sources."""

    def __init__(self, ev) -> None:
        """Initialize the Fetch class."""
        # Now we have access to the Evaluation object.
        self.ev = ev
        self.usgs_cache_dir = Path(
            ev.cache_dir,
            const.FETCHING_CACHE_DIR,
            const.USGS_CACHE_DIR,
        )
        self.nwm_cache_dir = Path(
            ev.cache_dir,
            const.FETCHING_CACHE_DIR,
            const.NWM_CACHE_DIR
        )
        self.kerchunk_cache_dir = Path(
            ev.cache_dir,
            const.FETCHING_CACHE_DIR,
            const.KERCHUNK_DIR
        )
        self.weights_cache_dir = Path(
            ev.cache_dir,
            const.FETCHING_CACHE_DIR,
            const.WEIGHTS_DIR
        )

    def _get_secondary_location_ids(self, prefix: str) -> List[str]:
        """Get the secondary location IDs corresponding to primary IDs."""
        lcw_df = self.ev.location_crosswalks.query(
            filters={
                "column": "secondary_location_id",
                "operator": "like",
                "value": f"{prefix}-%"
            }
        ).to_pandas()

        if lcw_df.empty:
            logger.error(
                "No secondary location IDs were found in the crosswalk table"
                f" with the specified prefix: '{prefix}'"
            )
            raise ValueError(
                "No secondary location IDs were found in the crosswalk table"
                f" with the specified prefix: '{prefix}'"
            )

        location_ids = (
            lcw_df.secondary_location_id.
            str.removeprefix(f"{prefix}-").to_list()
        )

        return location_ids

    def _configuration_name_exists(
        self, configuration_name: str
    ) -> bool:
        """Check if a configuration name exists in the configurations table.

        True: Configuration name exists in the table.
        False: Configuration name does not exist in the table.
        """
        sdf = self.ev.configurations.filter(
            {
                "column": "name",
                "operator": "=",
                "value": configuration_name
            }
        ).to_sdf()
        return not sdf.rdd.isEmpty()

    def usgs_streamflow(
        self,
        start_date: Union[str, datetime, pd.Timestamp],
        end_date: Union[str, datetime, pd.Timestamp],
        service: USGSServiceEnum = "iv",
        chunk_by: Union[USGSChunkByEnum, None] = None,
        filter_to_hourly: bool = True,
        filter_no_data: bool = True,
        convert_to_si: bool = True,
        overwrite_output: Optional[bool] = False,
        timeseries_type: TimeseriesTypeEnum = "primary",
        write_mode: TableWriteEnum = "append",
    ):
        """Fetch USGS gage data and load into the TEEHR dataset.

        Data is fetched for all USGS IDs in the locations table, and all
        dates and times within the files and in the cached file names are
        in UTC.

        Parameters
        ----------
        start_date : Union[str, datetime, pd.Timestamp]
            Start time of data to fetch.
        end_date : Union[str, datetime, pd.Timestamp]
            End time of data to fetch. Note, since start_date is inclusive for
            the USGS service, we subtract 1 minute from this time so we don't
            get overlap between consecutive calls.
        service : USGSServiceEnum, default = "iv"
            The USGS service to use for fetching data ('iv' for hourly
            instantaneous values or 'dv' for daily mean values).
        chunk_by : Union[str, None], default = None
            A string specifying how much data is fetched and read into
            memory at once. The default is to fetch all locations and all
            dates at once.
            Valid options =
            ["location_id", "day", "week", "month", "year", None].
        filter_to_hourly : bool = True
            Return only values that fall on the hour
            (i.e. drop 15 minute data).
        filter_no_data : bool = True
            Filter out -999 values.
        convert_to_si : bool = True
            Multiplies values by 0.3048**3 and sets ``measurement_units`` to
            "m3/s".
        overwrite_output : bool
            Flag specifying whether or not to overwrite output files if they
            already exist.  True = overwrite; False = fail.
        timeseries_type : str
            Whether to consider as the "primary" or "secondary" timeseries.
            Default is "primary".
        write_mode : TableWriteEnum, optional (default: "append")
            The write mode for the table. Options are "append" or "upsert".
            If "append", the Evaluation table will be appended with new data
            that does not already exist.
            If "upsert", existing data will be replaced and new data that
            does not exist will be appended.


        .. note::

            Data in the cache is cleared before each call to the fetch method. So if a
            long-running fetch is interrupted before the data is automatically loaded
            into the Evaluation, it should be loaded or cached manually. This will
            prevent it from being deleted when the fetch job is resumed.

        Examples
        --------
        Here we fetch over a year of USGS hourly streamflow data.
        Initially the data is saved to the cache directory, then it is
        validated and loaded into the TEEHR dataset.

        >>> import teehr
        >>> ev = teehr.Evaluation()

        Fetch the data for locations in the locations table.

        >>> eval.fetch.usgs_streamflow(
        >>>     start_date=datetime(2021, 1, 1),
        >>>     end_date=datetime(2022, 2, 28)
        >>> )

        .. note::

            USGS data can also be fetched outside of a TEEHR Evaluation
            by calling the method directly.

        >>> from teehr.fetching.usgs.usgs import usgs_to_parquet

        This requires specifying a list of USGS gage IDs and an output
        directory in addition to the above arguments.

        >>> usgs_to_parquet(
        >>>     sites=["02449838", "02450825"],
        >>>     start_date=datetime(2023, 2, 20),
        >>>     end_date=datetime(2023, 2, 25),
        >>>     output_parquet_dir=Path(Path().home(), "temp", "usgs"),
        >>>     chunk_by="day",
        >>>     overwrite_output=True
        >>> )
        """  # noqa
        logger.info("Getting primary location IDs.")
        locations_df = self.ev.locations.query(
            filters={
                "column": "id",
                "operator": "like",
                "value": f"usgs-%"
            }
        ).to_pandas()
        sites = locations_df["id"].str.removeprefix(f"usgs-").to_list()

        usgs_variable_name = USGS_VARIABLE_MAPPER[VARIABLE_NAME][service]

        # Clear out cache
        remove_dir_if_exists(self.usgs_cache_dir)

        usgs_to_parquet(
            sites=sites,
            start_date=start_date,
            end_date=end_date,
            output_parquet_dir=Path(
                self.usgs_cache_dir,
                USGS_CONFIGURATION_NAME,
                usgs_variable_name
            ),
            chunk_by=chunk_by,
            filter_to_hourly=filter_to_hourly,
            filter_no_data=filter_no_data,
            convert_to_si=convert_to_si,
            overwrite_output=overwrite_output,
            timeseries_type=timeseries_type
        )

        if (
            not self._configuration_name_exists(USGS_CONFIGURATION_NAME)
        ):
            self.ev.configurations.add(
                Configuration(
                    name=USGS_CONFIGURATION_NAME,
                    type="primary",
                    description="USGS Observations"
                )
            )

        validate_and_insert_timeseries(
            ev=self.ev,
            in_path=Path(
                self.usgs_cache_dir
            ),
            timeseries_type=timeseries_type,
            write_mode=write_mode
        )

    def nwm_retrospective_points(
        self,
        nwm_version: SupportedNWMRetroVersionsEnum,
        variable_name: ChannelRtRetroVariableEnum,
        start_date: Union[str, datetime, pd.Timestamp],
        end_date: Union[str, datetime, pd.Timestamp],
        chunk_by: Union[NWMChunkByEnum, None] = None,
        overwrite_output: Optional[bool] = False,
        domain: Optional[SupportedNWMRetroDomainsEnum] = "CONUS",
        timeseries_type: TimeseriesTypeEnum = "secondary",
        write_mode: TableWriteEnum = "append"
    ):
        """Fetch NWM retrospective point data and load into the TEEHR dataset.

        Data is fetched for all secondary location IDs in the locations
        crosswalk table that are prefixed by the NWM version, and all dates
        and times within the files and in the cache file names are in UTC.

        Parameters
        ----------
        nwm_version : SupportedNWMRetroVersionsEnum
            NWM retrospective version to fetch.
            Currently `nwm20`, `nwm21`, and `nwm30` supported.
            Note that since there is no change in NWM configuration between
            version 2.1 and 2.2, no retrospective run was produced for v2.2.
        variable_name : str
            Name of the NWM data variable to download.
            (e.g., "streamflow", "velocity", ...).
        start_date : Union[str, datetime, pd.Timestamp]
            Date to begin data ingest.
            Str formats can include YYYY-MM-DD or MM/DD/YYYY
            Rounds down to beginning of day.

            - v2.0: 1993-01-01
            - v2.1: 1979-01-01
            - v3.0: 1979-02-01
        end_date : Union[str, datetime, pd.Timestamp],
            Last date to fetch.  Rounds up to end of day.
            Str formats can include YYYY-MM-DD or MM/DD/YYYY.

            - v2.0: 2018-12-31
            - v2.1: 2020-12-31
            - v3.0: 2023-01-31
        chunk_by : Union[NWMChunkByEnum, None] = None,
            If None (default) saves all timeseries to a single file, otherwise
            the data is processed using the specified parameter.
            Can be: 'week', 'month', or 'year'.
        overwrite_output : bool = False,
            Whether output should overwrite files if they exist.
            Default is False.
        domain : str = "CONUS"
            Geographical domain when NWM version is v3.0.
            Acceptable values are "Alaska", "CONUS" (default), "Hawaii",
            and "PR". Only relevant when NWM version equals `nwm30`.
        timeseries_type : str
            Whether to consider as the "primary" or "secondary" timeseries.
            Default is "primary".
        write_mode : TableWriteEnum, optional (default: "append")
            The write mode for the table. Options are "append" or "upsert".
            If "append", the Evaluation table will be appended with new data
            that does not already exist.
            If "upsert", existing data will be replaced and new data that
            does not exist will be appended.


        .. note::

            Data in the cache is cleared before each call to the fetch method. So if a
            long-running fetch is interrupted before the data is automatically loaded
            into the Evaluation, it should be loaded or cached manually. This will
            prevent it from being deleted when the fetch job is resumed.

        Examples
        --------
        Here we fetch one days worth of NWM hourly streamflow data. Initially
        the data is saved to the cache directory, then it is validated and
        loaded into the TEEHR dataset.

        >>> import teehr
        >>> ev = teehr.Evaluation()

        >>> ev.fetch.nwm_retrospective_points(
        >>>     nwm_version="nwm30",
        >>>     variable_name="streamflow",
        >>>     start_date=datetime(2000, 1, 1),
        >>>     end_date=datetime(2000, 1, 2, 23)
        >>> )

        .. note::

            NWM data can also be fetched outside of a TEEHR Evaluation
            by calling the method directly.

        >>> import teehr.fetching.nwm.retrospective_points as nwm_retro

        Fetch and format the data, writing to the specified directory.

        >>> nwm_retro.nwm_retro_to_parquet(
        >>>     nwm_version="nwm20",
        >>>     variable_name="streamflow",
        >>>     start_date=datetime(2000, 1, 1),
        >>>     end_date=datetime(2000, 1, 2, 23),
        >>>     location_ids=[7086109, 7040481],
        >>>     output_parquet_dir=Path(Path.home(), "nwm20_retrospective")
        >>> )

        See Also
        --------
        :func:`teehr.fetching.nwm.retrospective_points.nwm_retro_to_parquet`
        """ # noqa
        ev_configuration_name = f"{nwm_version}_retrospective"
        ev_variable_name = format_nwm_variable_name(variable_name)

        logger.info("Getting secondary location IDs.")
        location_ids = self._get_secondary_location_ids(
            prefix=nwm_version
        )

        # Clear out cache
        remove_dir_if_exists(self.nwm_cache_dir)

        nwm_retro_to_parquet(
            nwm_version=nwm_version,
            variable_name=variable_name,
            start_date=start_date,
            end_date=end_date,
            location_ids=location_ids,
            output_parquet_dir=Path(
                self.nwm_cache_dir,
                ev_configuration_name,
                ev_variable_name
            ),
            chunk_by=chunk_by,
            overwrite_output=overwrite_output,
            domain=domain,
            variable_mapper=NWM_VARIABLE_MAPPER,
            timeseries_type=timeseries_type
        )

        if (
            not self._configuration_name_exists(ev_configuration_name)
        ):
            self.ev.configurations.add(
                Configuration(
                    name=ev_configuration_name,
                    type=timeseries_type,
                    description=f"{nwm_version} retrospective"
                )
            )

        validate_and_insert_timeseries(
            ev=self.ev,
            in_path=Path(
                self.nwm_cache_dir
            ),
            timeseries_type=timeseries_type,
            write_mode=write_mode
        )

    def nwm_retrospective_grids(
        self,
        nwm_version: SupportedNWMRetroVersionsEnum,
        variable_name: ForcingVariablesEnum,
        start_date: Union[str, datetime, pd.Timestamp],
        end_date: Union[str, datetime, pd.Timestamp],
        calculate_zonal_weights: bool = True,
        overwrite_output: Optional[bool] = False,
        chunk_by: Optional[NWMChunkByEnum] = None,
        domain: Optional[SupportedNWMRetroDomainsEnum] = "CONUS",
        location_id_prefix: Optional[str] = None,
        timeseries_type: TimeseriesTypeEnum = "primary",
        write_mode: TableWriteEnum = "append",
        zonal_weights_filepath: Optional[Union[Path, str]] = None
    ):
        """
        Fetch NWM retrospective gridded data, calculate zonal statistics (currently only
        mean is available) of selected variable for given zones, and load
        into the TEEHR dataset.

        Data is fetched for the location IDs in the locations
        table having a given location_id_prefix. All dates and times within the files
        and in the cache file names are in UTC.

        The zonal weights file, which contains the fraction each grid pixel overlaps each
        zone is necessary, and can be calculated and saved to the cache directory
        if it does not already exist.

        Parameters
        ----------
        nwm_version : SupportedNWMRetroVersionsEnum
            NWM retrospective version to fetch.
            Currently `nwm21` and `nwm30` supported.
            Note that since there is no change in NWM configuration between
            version 2.1 and 2.2, no retrospective run was produced for v2.2.
        variable_name : str
            Name of the NWM forcing data variable to download.
            (e.g., "PRECIP", "PSFC", "Q2D", ...).
        start_date : Union[str, datetime, pd.Timestamp]
            Date to begin data ingest.
            Str formats can include YYYY-MM-DD or MM/DD/YYYY.
            Rounds down to beginning of day.

            - v2.0: 1993-01-01
            - v2.1: 1979-01-01
            - v3.0: 1979-02-01
        end_date : Union[str, datetime, pd.Timestamp],
            Last date to fetch.  Rounds up to end of day.
            Str formats can include YYYY-MM-DD or MM/DD/YYYY.

            - v2.0: 2018-12-31
            - v2.1: 2020-12-31
            - v3.0: 2023-01-31
        calculate_zonal_weights : bool
            Flag specifying whether or not to calculate zonal weights.
            True = calculate; False = use existing file. Default is True.
        location_id_prefix : Optional[str]
            Prefix to include when filtering the locations table for polygon
            primary_location_id. Default is None, all locations are included.
        overwrite_output : bool
            Flag specifying whether or not to overwrite output files if they already
            exist.  True = overwrite; False = fail.
        chunk_by : Optional[NWMChunkByEnum] = None,
            If None (default) saves all timeseries to a single file, otherwise
            the data is processed using the specified parameter.
            Can be: 'week' or 'month' for gridded data.
        domain : str = "CONUS"
            Geographical domain when NWM version is v3.0.
            Acceptable values are "Alaska", "CONUS" (default), "Hawaii",
            and "PR". Only relevant when NWM version equals v3.0.
        timeseries_type : str
            Whether to consider as the "primary" or "secondary" timeseries.
            Default is "primary".
        zonal_weights_filepath : Optional[Union[Path, str]]
            The path to the zonal weights file. If None and calculate_zonal_weights
            is False, the weights file must exist in the cache for the configuration.
            Default is None.

        Examples
        --------
        Here we will calculate mean areal precipitation using NWM forcing data for
        the polygons in the locations table. Pixel weights (fraction of pixel overlap)
        are calculated for each polygon and stored in the evaluation cache directory.

        (see: :func:`generate_weights_file()
        <teehr.utilities.generate_weights.generate_weights_file>` for weights calculation).

        >>> import teehr
        >>> ev = teehr.Evaluation()

        >>> ev.fetch.nwm_retrospective_grids(
        >>>     nwm_version="nwm30",
        >>>     variable_name="RAINRATE",
        >>>     calculate_zonal_weights=True,
        >>>     start_date=datetime(2000, 1, 1),
        >>>     end_date=datetime(2001, 1, 1),
        >>>     location_id_prefix="huc10"
        >>> )

        .. note::

            NWM data can also be fetched outside of a TEEHR Evaluation
            by calling the method directly.

        >>> from teehr.fetching.nwm.retrospective_grids import nwm_retro_grids_to_parquet

        Perform the calculations, writing to the specified directory.

        >>> nwm_retro_grids_to_parquet(
        >>>     nwm_version="nwm30",
        >>>     variable_name="RAINRATE",
        >>>     zonal_weights_filepath=Path(Path.home(), "nextgen_03S_weights.parquet"),
        >>>     start_date=2020-12-18,
        >>>     end_date=2022-12-18,
        >>>     output_parquet_dir=Path(Path.home(), "temp/parquet"),
        >>>     location_id_prefix="huc10",
        >>> )

        See Also
        --------
        :func:`teehr.fetching.nwm.nwm_grids.nwm_grids_to_parquet`
        """ # noqa
        ev_configuration_name = f"{nwm_version}_retrospective"
        ev_variable_name = format_nwm_variable_name(variable_name)

        # Clear out cache
        remove_dir_if_exists(self.nwm_cache_dir)
        ev_weights_cache_dir = Path(
            self.weights_cache_dir, ev_configuration_name
        )
        ev_weights_cache_dir.mkdir(parents=True, exist_ok=True)

        # Note: If the weights file will be generated, use the location ID
        # prefix to filter the locations table for the correct locations,
        # defining the zone_polygons argument.
        logger.info("Getting primary location IDs.")
        if location_id_prefix is None:
            locations_gdf = self.ev.locations.to_geopandas()
        else:
            locations_gdf = self.ev.locations.query(
                filters={
                    "column": "id",
                    "operator": "like",
                    "value": f"{location_id_prefix}-%"
                }
            ).to_geopandas()

        if zonal_weights_filepath is None:
            zonal_weights_filepath = Path(
                ev_weights_cache_dir,
                f"{ev_configuration_name}_pixel_weights.parquet"
            )

        nwm_retro_grids_to_parquet(
            nwm_version=nwm_version,
            variable_name=variable_name,
            zonal_weights_filepath=zonal_weights_filepath,
            start_date=start_date,
            end_date=end_date,
            output_parquet_dir=Path(
                self.nwm_cache_dir,
                ev_configuration_name,
                ev_variable_name
            ),
            chunk_by=chunk_by,
            overwrite_output=overwrite_output,
            domain=domain,
            location_id_prefix=location_id_prefix,
            variable_mapper=NWM_VARIABLE_MAPPER,
            timeseries_type=timeseries_type,
            unique_zone_id="id",
            calculate_zonal_weights=calculate_zonal_weights,
            zone_polygons=locations_gdf
        )

        if (
            not self._configuration_name_exists(ev_configuration_name)
        ):
            self.ev.configurations.add(
                Configuration(
                    name=ev_configuration_name,
                    type=timeseries_type,
                    description=f"{nwm_version} retrospective"
                )
            )

        validate_and_insert_timeseries(
            ev=self.ev,
            in_path=Path(
                self.nwm_cache_dir
            ),
            timeseries_type=timeseries_type,
            write_mode=write_mode
        )

    def nwm_operational_points(
        self,
        nwm_configuration: str,
        output_type: str,
        variable_name: str,
        start_date: Union[str, datetime],
        ingest_days: int,
        nwm_version: SupportedNWMOperationalVersionsEnum,
        data_source: Optional[SupportedNWMDataSourcesEnum] = "GCS",
        kerchunk_method: Optional[SupportedKerchunkMethod] = "local",
        prioritize_analysis_valid_time: Optional[bool] = False,
        t_minus_hours: Optional[List[int]] = None,
        process_by_z_hour: Optional[bool] = True,
        stepsize: Optional[int] = 100,
        ignore_missing_file: Optional[bool] = True,
        overwrite_output: Optional[bool] = False,
        timeseries_type: TimeseriesTypeEnum = "secondary",
        starting_z_hour: Optional[int] = None,
        ending_z_hour: Optional[int] = None,
        write_mode: TableWriteEnum = "append"
    ):
        """Fetch operational NWM point data and load into the TEEHR dataset.

        Data is fetched for all secondary location IDs in the locations
        crosswalk table that are prefixed by the NWM version, and all dates
        and times within the files and in the cache file names are in UTC.

        Parameters
        ----------
        nwm_configuration : str
            NWM forecast category.
            (e.g., "analysis_assim", "short_range", ...).
        output_type : str
            Output component of the nwm_configuration.
            (e.g., "channel_rt", "reservoir", ...).
        variable_name : str
            Name of the NWM data variable to download.
            (e.g., "streamflow", "velocity", ...).
        start_date : str or datetime
            Date to begin data ingest.
            Str formats can include YYYY-MM-DD or MM/DD/YYYY.
        ingest_days : int
            Number of days to ingest data after start date.
        nwm_version : SupportedNWMOperationalVersionsEnum
            The NWM operational version.
            "nwm12", "nwm20", "nwm21", "nwm22", or "nwm30".
            Note that there is no change in NWM configuration between
            version 2.1 and 2.2, and they are treated as the same version.
            They are both allowed here for convenience.

            Availability of each version:

            - v1.2: 2018-09-17 - 2019-06-18
            - v2.0: 2019-06-19 - 2021-04-19
            - v2.1/2.2: 2021-04-20 - 2023-09-18
            - v3.0: 2023-09-19 - present
        data_source : Optional[SupportedNWMDataSourcesEnum]
            Specifies the remote location from which to fetch the data
            "GCS" (default), "NOMADS", or "DSTOR"
            Currently only "GCS" is implemented.
        kerchunk_method : Optional[SupportedKerchunkMethod]
            When data_source = "GCS", specifies the preference in creating Kerchunk
            reference json files. "local" (default) will create new json files from
            netcdf files in GCS and save to a local directory if they do not already
            exist locally, in which case the creation is skipped. "remote" - read the
            CIROH pre-generated jsons from s3, ignoring any that are unavailable.
            "auto" - read the CIROH pre-generated jsons from s3, and create any that
            are unavailable, storing locally.
        prioritize_analysis_valid_time : Optional[bool]
            A boolean flag that determines the method of fetching analysis data.
            When False (default), all hours of the reference time are included in the
            output. When True, only the hours within t_minus_hours are included.
        t_minus_hours : Optional[List[int]]
            Specifies the look-back hours to include if an assimilation
            nwm_configuration is specified.
        process_by_z_hour : Optional[bool]
            A boolean flag that determines the method of grouping files
            for processing. The default is True, which groups by day and z_hour.
            False groups files sequentially into chunks, whose size is determined
            by stepsize. This allows users to process more data potentially more
            efficiently, but runs to risk of splitting up forecasts into separate
            output files.
        stepsize : Optional[int]
            The number of json files to process at one time. Used if
            process_by_z_hour is set to False. Default value is 100. Larger values
            can result in greater efficiency but require more memory.
        ignore_missing_file : Optional[bool]
            Flag specifying whether or not to fail if a missing NWM file is
            encountered. True = skip and continue; False = fail.
        overwrite_output : Optional[bool]
            Flag specifying whether or not to overwrite output files if they
            already exist.  True = overwrite; False = fail.
        timeseries_type : str
            Whether to consider as the "primary" or "secondary" timeseries.
            Default is "secondary".
        starting_z_hour : Optional[int]
            The starting z_hour to include in the output. If None, all z_hours
            are included for the first day. Default is None. Must be between 0 and 23.
        ending_z_hour : Optional[int]
            The ending z_hour to include in the output. If None, all z_hours
            are included for the last day. Default is None. Must be between 0 and 23.
        write_mode : TableWriteEnum, optional (default: "append")
            The write mode for the table. Options are "append" or "upsert".
            If "append", the Evaluation table will be appended with new data
            that does not already exist.
            If "upsert", existing data will be replaced and new data that
            does not exist will be appended.


        .. note::

            Data in the cache is cleared before each call to the fetch method. So if a
            long-running fetch is interrupted before the data is automatically loaded
            into the Evaluation, it should be loaded or cached manually. This will
            prevent it from being deleted when the fetch job is resumed.

        Notes
        -----
        The NWM variables, including nwm_configuration, output_type, and
        variable_name are stored as pydantic models in point_config_models.py

        The cached forecast and assimilation data is grouped and saved one file
        per reference time, using the file name convention "YYYYMMDDTHH".

        Examples
        --------
        Here we fetch operational streamflow forecasts for NWM v2.2 from GCS, and
        load into the TEEHR dataset.

        >>> import teehr
        >>> ev = teehr.Evaluation()

        >>> ev.fetch.nwm_operational_points(
        >>>     nwm_configuration="short_range",
        >>>     output_type="channel_rt",
        >>>     variable_name="streamflow",
        >>>     start_date=datetime(2000, 1, 1),
        >>>     ingest_days=1,
        >>>     nwm_version="nwm21",
        >>>     data_source="GCS",
        >>>     kerchunk_method="auto"
        >>> )

        .. note::

            NWM data can also be fetched outside of a TEEHR Evaluation
            by calling the method directly.

        >>> from teehr.fetching.nwm.nwm_points import nwm_to_parquet

        Fetch and format the data, writing to the specified directory.

        >>> nwm_to_parquet(
        >>>     nwm_configuration="short_range",
        >>>     output_type="channel_rt",
        >>>     variable_name="streamflow",
        >>>     start_date="2023-03-18",
        >>>     ingest_days=1,
        >>>     location_ids=LOCATION_IDS,
        >>>     json_dir=Path(Path.home(), "temp/parquet/jsons/"),
        >>>     output_parquet_dir=Path(Path.home(), "temp/parquet"),
        >>>     nwm_version="nwm21",
        >>>     data_source="GCS",
        >>>     kerchunk_method="auto",
        >>>     prioritize_analysis_valid_time=True,
        >>>     t_minus_hours=[0, 1, 2],
        >>>     process_by_z_hour=True,
        >>>     stepsize=STEPSIZE,
        >>>     ignore_missing_file=True,
        >>>     overwrite_output=True,
        >>> )

        See Also
        --------
        :func:`teehr.fetching.nwm.nwm_points.nwm_to_parquet`
        """ # noqa
        logger.info("Getting secondary location IDs.")

        location_ids = self._get_secondary_location_ids(
            prefix=nwm_version
        )

        ev_variable_name = format_nwm_variable_name(variable_name)
        ev_config = format_nwm_configuration_name(
            nwm_configuration_name=nwm_configuration,
            nwm_version=nwm_version
        )

        # Clear out cache
        remove_dir_if_exists(self.nwm_cache_dir)

        nwm_to_parquet(
            configuration=nwm_configuration,
            output_type=output_type,
            variable_name=variable_name,
            start_date=start_date,
            ingest_days=ingest_days,
            location_ids=location_ids,
            json_dir=self.kerchunk_cache_dir,
            output_parquet_dir=Path(
                self.nwm_cache_dir,
                ev_config["configuration_name"],
                ev_variable_name
            ),
            nwm_version=nwm_version,
            data_source=data_source,
            kerchunk_method=kerchunk_method,
            prioritize_analysis_valid_time=prioritize_analysis_valid_time,
            t_minus_hours=t_minus_hours,
            process_by_z_hour=process_by_z_hour,
            stepsize=stepsize,
            ignore_missing_file=ignore_missing_file,
            overwrite_output=overwrite_output,
            variable_mapper=NWM_VARIABLE_MAPPER,
            timeseries_type=timeseries_type,
            starting_z_hour=starting_z_hour,
            ending_z_hour=ending_z_hour
        )

        if (
            not self._configuration_name_exists(
                ev_config["configuration_name"]
            )
        ):
            self.ev.configurations.add(
                Configuration(
                    name=ev_config["configuration_name"],
                    type=timeseries_type,
                    description=f"{nwm_version} operational forecasts"
                )
            )

        validate_and_insert_timeseries(
            ev=self.ev,
            in_path=Path(
                self.nwm_cache_dir
            ),
            timeseries_type=timeseries_type,
            write_mode=write_mode
        )

    def nwm_operational_grids(
        self,
        nwm_configuration: str,
        output_type: str,
        variable_name: str,
        start_date: Union[str, datetime],
        ingest_days: int,
        nwm_version: SupportedNWMOperationalVersionsEnum,
        calculate_zonal_weights: bool = True,
        location_id_prefix: Optional[str] = None,
        data_source: Optional[SupportedNWMDataSourcesEnum] = "GCS",
        kerchunk_method: Optional[SupportedKerchunkMethod] = "local",
        prioritize_analysis_valid_time: Optional[bool] = False,
        t_minus_hours: Optional[List[int]] = None,
        ignore_missing_file: Optional[bool] = True,
        overwrite_output: Optional[bool] = False,
        timeseries_type: TimeseriesTypeEnum = "primary",
        starting_z_hour: Optional[int] = None,
        ending_z_hour: Optional[int] = None,
        write_mode: TableWriteEnum = "append",
        zonal_weights_filepath: Optional[Union[Path, str]] = None,
    ):
        """
        Fetch NWM operational gridded data, calculate zonal statistics (currently only
        mean is available) of selected variable for given zones, and load into
        the TEEHR dataset.

        Data is fetched for the location IDs in the locations
        table having a given location_id_prefix. All dates and times within the files
        and in the cache file names are in UTC.

        The zonal weights file, which contains the fraction each grid pixel overlaps each
        zone is necessary, and can be calculated and saved to the cache directory
        if it does not already exist.

        Parameters
        ----------
        nwm_configuration : str
            NWM forecast category.
            (e.g., "analysis_assim", "short_range", ...).
        output_type : str
            Output component of the nwm_configuration.
            (e.g., "channel_rt", "reservoir", ...).
        variable_name : str
            Name of the NWM data variable to download.
            (e.g., "streamflow", "velocity", ...).
        start_date : str or datetime
            Date to begin data ingest.
            Str formats can include YYYY-MM-DD or MM/DD/YYYY.
        ingest_days : int
            Number of days to ingest data after start date.
        nwm_version : SupportedNWMOperationalVersionsEnum
            The NWM operational version.
            "nwm12", "nwm20", "nwm21", "nwm22", or "nwm30".
            Note that there is no change in NWM configuration between
            version 2.1 and 2.2, and they are treated as the same version.
            They are both allowed here for convenience.

            Availability of each version:

            - v1.2: 2018-09-17 - 2019-06-18
            - v2.0: 2019-06-19 - 2021-04-19
            - v2.1/2.2: 2021-04-20 - 2023-09-18
            - v3.0: 2023-09-19 - present
        calculate_zonal_weights : bool
            Flag specifying whether or not to calculate zonal weights.
            True = calculate; False = use existing file. Default is True.
        location_id_prefix : Optional[str]
            Prefix to include when filtering the locations table for polygon
            primary_location_id. Default is None, all locations are included.
        data_source : Optional[SupportedNWMDataSourcesEnum]
            Specifies the remote location from which to fetch the data
            "GCS" (default), "NOMADS", or "DSTOR".
            Currently only "GCS" is implemented.
        kerchunk_method : Optional[SupportedKerchunkMethod]
            When data_source = "GCS", specifies the preference in creating Kerchunk
            reference json files. "local" (default) will create new json files from
            netcdf files in GCS and save to a local directory if they do not already
            exist locally, in which case the creation is skipped. "remote" - read the
            CIROH pre-generated jsons from s3, ignoring any that are unavailable.
            "auto" - read the CIROH pre-generated jsons from s3, and create any that
            are unavailable, storing locally.
        prioritize_analysis_valid_time : Optional[bool]
            A boolean flag that determines the method of fetching analysis data.
            When False (default), all hours of the reference time are included in the
            output. When True, only the hours within t_minus_hours are included.
        t_minus_hours : Optional[Iterable[int]]
            Specifies the look-back hours to include if an assimilation
            nwm_configuration is specified.
        ignore_missing_file : bool
            Flag specifying whether or not to fail if a missing NWM file is encountered
            True = skip and continue; False = fail.
        overwrite_output : bool
            Flag specifying whether or not to overwrite output files if they already
            exist.  True = overwrite; False = fail.
        timeseries_type : str
            Whether to consider as the "primary" or "secondary" timeseries.
            Default is "primary".
        starting_z_hour : Optional[int]
            The starting z_hour to include in the output. If None, all z_hours
            are included for the first day. Default is None. Must be between 0 and 23.
        ending_z_hour : Optional[int]
            The ending z_hour to include in the output. If None, all z_hours
            are included for the last day. Default is None. Must be between 0 and 23.
        write_mode : TableWriteEnum, optional (default: "append")
            The write mode for the table. Options are "append" or "upsert".
            If "append", the Evaluation table will be appended with new data
            that does not already exist.
            If "upsert", existing data will be replaced and new data that
            does not exist will be appended.
        zonal_weights_filepath : Optional[Union[Path, str]]
            The path to the zonal weights file. If None and calculate_zonal_weights
            is False, the weights file must exist in the cache for the configuration.
            Default is None.


        .. note::

            Data in the cache is cleared before each call to the fetch method. So if a
            long-running fetch is interrupted before the data is automatically loaded
            into the Evaluation, it should be loaded or cached manually. This will
            prevent it from being deleted when the fetch job is resumed.

        Notes
        -----
        The NWM variables, including nwm_configuration, output_type, and
        variable_name are stored as a pydantic model in grid_config_models.py.

        The cached forecast and assimilation data is grouped and saved one file
        per reference time, using the file name convention "YYYYMMDDTHH".

        All dates and times within the files and in the file names are in UTC.

        Examples
        --------
        Here we will calculate mean areal precipitation using operational NWM forcing data
        for the polygons in the locations table. Pixel weights (fraction of pixel overlap)
        are calculated for each polygon and stored in the evaluation cache directory.

        (see: :func:`generate_weights_file()
        <teehr.utilities.generate_weights.generate_weights_file>` for weights calculation).

        >>> import teehr
        >>> ev = teehr.Evaluation()

        >>> ev.fetch.nwm_operational_grids(
        >>>     nwm_configuration="forcing_short_range",
        >>>     output_type="forcing",
        >>>     variable_name="RAINRATE",
        >>>     start_date=datetime(2000, 1, 1),
        >>>     ingest_days=1,
        >>>     Path(Path.home(), "nextgen_03S_weights.parquet"),
        >>>     nwm_version="nwm22",
        >>>     data_source="GCS",
        >>>     kerchunk_method="auto"
        >>> )

        .. note::

            NWM data can also be fetched outside of a TEEHR Evaluation
            by calling the method directly.

        >>> from teehr.fetching.nwm.nwm_grids import nwm_grids_to_parquet

        Perform the calculations, writing to the specified directory.

        >>> nwm_grids_to_parquet(
        >>>     nwm_configuration=forcing_short_range,
        >>>     output_type=forcing,
        >>>     variable_name=RAINRATE,
        >>>     start_date=2020-12-18,
        >>>     ingest_days=1,
        >>>     zonal_weights_filepath=Path(Path.home(), "nextgen_03S_weights.parquet"),
        >>>     json_dir=Path(Path.home(), "temp/parquet/jsons/"),
        >>>     output_parquet_dir=Path(Path.home(), "temp/parquet"),
        >>>     nwm_version="nwm21",
        >>>     data_source="GCS",
        >>>     kerchunk_method="auto",
        >>>     t_minus_hours=[0, 1, 2],
        >>>     ignore_missing_file=True,
        >>>     overwrite_output=True
        >>> )

        See Also
        --------
        :func:`teehr.fetching.nwm.nwm_grids.nwm_grids_to_parquet`
        """ # noqa
        ev_variable_name = format_nwm_variable_name(variable_name)
        ev_config = format_nwm_configuration_name(
            nwm_configuration_name=nwm_configuration,
            nwm_version=nwm_version
        )

        # Clear out cache
        remove_dir_if_exists(self.nwm_cache_dir)

        ev_weights_cache_dir = Path(
            self.weights_cache_dir, ev_config["configuration_name"]
        )
        ev_weights_cache_dir.mkdir(parents=True, exist_ok=True)

        # Note: If the weights file will be generated, use the location ID
        # prefix to filter the locations table for the correct locations,
        # defining the zone_polygons argument.
        logger.info("Getting primary location IDs.")
        if location_id_prefix is None:
            locations_gdf = self.ev.locations.to_geopandas()
        else:
            locations_gdf = self.ev.locations.query(
                filters={
                    "column": "id",
                    "operator": "like",
                    "value": f"{location_id_prefix}-%"
                }
            ).to_geopandas()

        if zonal_weights_filepath is None:
            zonal_weights_filepath = Path(
                ev_weights_cache_dir,
                f"{ev_config['configuration_name']}_pixel_weights.parquet"
            )

        nwm_grids_to_parquet(
            configuration=nwm_configuration,
            output_type=output_type,
            variable_name=variable_name,
            start_date=start_date,
            ingest_days=ingest_days,
            zonal_weights_filepath=zonal_weights_filepath,
            json_dir=self.kerchunk_cache_dir,
            output_parquet_dir=Path(
                self.nwm_cache_dir,
                ev_config["configuration_name"],
                ev_variable_name
            ),
            nwm_version=nwm_version,
            data_source=data_source,
            kerchunk_method=kerchunk_method,
            prioritize_analysis_valid_time=prioritize_analysis_valid_time,
            t_minus_hours=t_minus_hours,
            ignore_missing_file=ignore_missing_file,
            overwrite_output=overwrite_output,
            location_id_prefix=location_id_prefix,
            variable_mapper=NWM_VARIABLE_MAPPER,
            starting_z_hour=starting_z_hour,
            ending_z_hour=ending_z_hour,
            unique_zone_id="id",
            calculate_zonal_weights=calculate_zonal_weights,
            zone_polygons=locations_gdf,
            timeseries_type=timeseries_type
        )

        if (
            not self._configuration_name_exists(
                ev_config["configuration_name"]
            )
        ):
            self.ev.configurations.add(
                Configuration(
                    name=ev_config["configuration_name"],
                    type=timeseries_type,
                    description=f"{nwm_version} operational forecasts"
                )
            )

        validate_and_insert_timeseries(
            ev=self.ev,
            in_path=Path(self.nwm_cache_dir),
            timeseries_type=timeseries_type,
            write_mode=write_mode
        )
